from collections.abc import MutableMapping
from dataclasses import dataclass, field
from typing import Any, List, Mapping


class MappingBase(MutableMapping):
    data: List[Mapping[str, Any]]

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key, value):
        del self.data[key]

    def __iter__(self):
        for k in self.data:
            yield k

    def __len__(self):
        return len(self.data)

    def as_dict(self):
        return self.data


@dataclass
class Row(MappingBase):
    """
    Row data container
    """

    row_id: str
    table_name: str
    data: List[Mapping[str, Any]]
    parent: object


@dataclass
class Rows(MappingBase):
    """
    Flattened rows for each object
    """

    ocid: str
    buyer: Mapping[str, str]
    data: Mapping[str, List[Row]] = field(default_factory=list)
    row: Row = ""

    def new_row(self, table, item_id):
        name = table.name
        head = self.ocid
        row = {
            "id": item_id,
            "ocid": self.ocid,
        }
        parent_row = self.row

        if table.is_combined:
            while parent_row and parent_row.parent:
                parent_row = parent_row.parent
            if parent_row:
                row["parentTable"] = parent_row.table_name

        if not table.is_root:
            parent_table = table.parent
            while parent_row and parent_row.parent:
                if parent_row.table_name == parent_table.name:
                    break
                if parent_row:
                    parent_row = parent_row.parent
            if parent_table:
                row["parentTable"] = parent_table.name
            if parent_row:
                parent_id = parent_row.row_id
                row["parentID"] = parent_id
                head = parent_id
        row_id = f"{head}/{name}:{item_id}"
        row["rowID"] = row_id
        if table.name == "parties" and self.buyer:
            # it works and we bind to OCDS anyway
            # but this is a hack
            # and its a good idea to find better way of doing it in future
            row["/buyer/id"] = self.buyer.get("id")
            row["/buyer/name"] = self.buyer.get("name")
        row = Row(row_id=row_id, data=row, table_name=table.name, parent=parent_row)
        self.row = row
        return row



import gettext
import locale
import warnings
from functools import lru_cache

import pkg_resources

DOMAIN = "spoonbill"
LOCALEDIR = pkg_resources.resource_filename(DOMAIN, "locale/")
LOCALE = "en"

system_locale = locale.getdefaultlocale()
if system_locale and system_locale[0]:
    LOCALE = system_locale[0]


def translate(msg_id, lang=LOCALE):
    """Simple wrapper of python's gettext with ability to override desired language"""
    try:
        translator = _translation(lang)
    except FileNotFoundError:
        LOCALE = "en"
        lang = LOCALE
        translator = _translation(lang)

    if translator:
        return translator.gettext(msg_id)
    return msg_id


@lru_cache(maxsize=None)
def _translation(lang):
    try:
        return gettext.translation(DOMAIN, LOCALEDIR, languages=[lang], fallback=None)
    except FileNotFoundError as e:
        warnings.warn(f"{e.strerror}: {e.filename}")
        return gettext.NullTranslations()


_ = translate



import csv

import openpyxl

from spoonbill.writers import CSVWriter, XlsxWriter


def read_csv_headers(path):
    with open(path, encoding="utf-8") as fd:
        return [c for c in csv.DictReader(fd).fieldnames]


def read_xlsx_headers(path, sheet):
    wb = openpyxl.load_workbook(filename=path)
    sheet = wb[sheet]
    columns = sheet.columns
    return [c.value for cell in columns for c in cell]


def get_writers(workdir, tables, options, schema=None):

    with CSVWriter(workdir, tables, options, schema=schema) as csv, XlsxWriter(
        workdir, tables, options, schema=schema
    ) as xlsx:
        return [csv, xlsx]


def prepare_tables(spec, options, inc_columns=None):
    tables = {name: table for name, table in spec.tables.items() if name in options.selection}
    if inc_columns:
        for name, table in tables.items():
            table.inc_column(inc_columns[name], inc_columns[name])
    return tables



import json
import pathlib
from decimal import Decimal

import pytest

from spoonbill.flatten import FlattenOptions
from spoonbill.spec import Column, Table
from spoonbill.stats import DataPreprocessor

from .data import TEST_COMBINED_TABLES, TEST_ROOT_TABLES

here = pathlib.Path(__file__).parent
schema_path = here / "data" / "ocds-simplified-schema.json"
releases_path = here / "data" / "ocds-sample-data.json"
test_dataset_path = here / "data" / "test_data.json"
dataset_path_with_combined_tables = here / "data" / "releases_with_combined_tables.json"
analyzed_path = here / "data" / "analyzed"
releases_extension_path = here / "data" / "ocds-sample-data-extension.json"


@pytest.fixture
def schema():
    with open(schema_path, encoding="utf-8") as fd:
        return json.load(fd, parse_float=Decimal)


@pytest.fixture
def releases():
    with open(releases_path, encoding="utf-8") as fd:
        return json.load(fd)["releases"]


@pytest.fixture
def releases_extension():
    with open(releases_extension_path, encoding="utf-8") as fd:
        return json.load(fd)["releases"]


@pytest.fixture
def flatten_options():
    return FlattenOptions(
        **{
            "selection": {"tenders": {"split": True}, "parties": {"split": False}},
        }
    )


@pytest.fixture
def spec(schema):
    return DataPreprocessor(schema, TEST_ROOT_TABLES, combined_tables=TEST_COMBINED_TABLES, with_preview=True)


@pytest.fixture
def spec_analyzed(schema, releases):
    dp = DataPreprocessor(schema, TEST_ROOT_TABLES, combined_tables=TEST_COMBINED_TABLES, with_preview=True)
    [_ for _ in dp.process_items(releases)]
    return dp


@pytest.fixture
def root_table():
    return Table(
        name="tenders",
        path=["/tender"],
        is_root=True,
        columns={
            "ocid": Column(id="ocid", path="ocid", title="ocid", type="string", hits=0),
            "id": Column(id="id", path="id", title="id", type="string", hits=0),
            "rowID": Column(path="rowID", title="rowID", type="string", id="rowID", hits=0),
            "parentID": Column(path="parentID", title="parentID", type="string", id="parentID", hits=0),
            "/tender/awardCriteriaDetails": Column(
                title="Tender Award Criteria Details",
                type="string",
                id="/tender/awardCriteriaDetails",
                path="/tender/awardCriteriaDetails",
                hits=0,
            ),
            "/tender/items/0/id": Column(
                title="Tender Item id",
                type="string",
                hits=0,
                id="/tender/items/0/id",
                path="/tender/items/0/id",
            ),
            "/tender/items/0/additionalClassifications/0/id": Column(
                title="Tender Item Classification id",
                type="string",
                id="/tender/items/0/additionalClassifications/0/id",
                path="/tender/items/0/additionalClassifications/0/id",
            ),
        },
        combined_columns={
            "ocid": Column(title="ocid", type="string", id="ocid", path="ocid", hits=0),
            "id": Column(title="id", path="id", type="string", id="id", hits=0),
            "rowID": Column(title="rowID", type="string", id="rowID", path="rowID", hits=0),
            "parentID": Column(title="parentID", type="string", id="parentID", path="parentID", hits=0),
            "/tender/awardCriteriaDetails": Column(
                title="Tender Award Criteria Details",
                type="string",
                id="/tender/awardCriteriaDetails",
                path="/tender/awardCriteriaDetails",
                hits=0,
            ),
            "/tender/submissionMethod": Column(
                title="Tender Submission Method",
                type="array",
                id="/tender/submissionMethod",
                path="/tender/submissionMethod",
            ),
            "/tender/items/0/id": Column(title="Tender Item id", type="string", id="/tender/items/0/id", path="id"),
            "/tender/items/0/additionalClassifications/0/id": Column(
                title="Tender Item Classification id",
                type="string",
                id="/tender/items/0/additionalClassifications/0/id",
                path="id",
            ),
        },
        arrays={"/tender/items": 0, "/tender/items/additionalClassifications": 0},
    )


@pytest.fixture
def test_dataset_releases():
    with open(test_dataset_path, encoding="utf-8") as fd:
        return json.load(fd)["releases"]


@pytest.fixture
def releases_with_combined_tables():
    with open(dataset_path_with_combined_tables, encoding="utf-8") as fd:
        return json.load(fd)



from collections import defaultdict

import pytest
from jmespath import search
from jsonpointer import resolve_pointer

from spoonbill.common import JOINABLE_SEPARATOR
from spoonbill.flatten import Flattener, FlattenOptions

ID_ITEMS = {
    "tenders": [
        {"/tender/id": "ocds-213czf-000-00001-01-planning"},
        {"/tender/id": "ocds-213czf-000-00001-01-tender"},
        {"/tender/id": "ocds-213czf-000-00001-01-tender"},
        {"/tender/id": "ocds-213czf-000-00001-01-tender"},
    ],
    "parties": [
        {"/parties/id": "GB-LAC-E09000003"},
        {"/parties/id": "GB-LAC-E09000003"},
        {"/parties/id": "GB-LAC-E09000003"},
        {"/parties/id": "GB-COH-22222222"},
        {"/parties/id": "GB-COH-11111111"},
        {"/parties/id": "GB-LAC-E09000003"},
        {"/parties/id": "GB-LAC-E09000003"},
        {"/parties/id": "GB-LAC-E09000003"},
    ],
}


def test_flatten(spec_analyzed, releases):
    options = FlattenOptions(
        **{
            "selection": {"tenders": {"split": True}, "parties": {"split": False}},
        }
    )
    flattener = Flattener(options, spec_analyzed.tables)
    count = {"tenders": 0, "parties": 0}
    for _count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            for row in rows:
                assert "id" in row
                assert "ocid" in row
                assert "rowID" in row
                if name in ID_ITEMS:
                    key = "tender" if name == "tenders" else "parties"
                    path = f"/{key}/id"
                    assert ID_ITEMS[name][count[name]][path] == row.get(path)
                    count[name] += 1


def test_flatten_row_id_parent_id_relation(spec, releases):
    releases[0]["tender"]["items"] = releases[0]["tender"]["items"] * 6
    releases[0]["tender"]["items"] = releases[0]["tender"]["items"] * 6
    releases[0]["tender"]["items"][0]["additionalClassifications"] = (
        releases[0]["tender"]["items"][0]["additionalClassifications"] * 6
    )
    for _ in spec.process_items(releases):
        pass
    options = FlattenOptions(**{"selection": {"tenders": {"split": True}}})
    flattener = Flattener(options, spec.tables)
    all_rows = defaultdict(list)
    for count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            all_rows[name].extend(rows)

    for row in all_rows["tenders_items_class"]:
        parent_id = row["parentID"]
        items = [i for i in all_rows["tenders_items"] if i["rowID"] == parent_id]
        assert items


def test_flattener_not_generate_count_columns(spec, releases):
    for _ in spec.process_items(releases):
        pass
    options = FlattenOptions(**{"selection": {"tenders": {"split": False}}, "count": True})
    flattener = Flattener(options, spec.tables)
    tenders = flattener.tables["tenders"]
    assert "/tender/itemsCount" not in tenders
    for index in range(tenders.arrays["/tender/items/additionalClassifications"]):
        assert f"/tender/items/{index}/additionalClassificationsCount" not in tenders

    options = FlattenOptions(
        **{"selection": {"tenders": {"split": True}, "tenders_items": {"split": False}}, "count": True}
    )
    flattener = Flattener(options, spec.tables)
    tenders = flattener.tables["tenders"]
    tenders_items = flattener.tables["tenders_items"]
    assert "/tender/itemsCount" not in tenders.combined_columns
    for index in range(tenders.arrays["/tender/items/additionalClassifications"]):
        assert f"/tender/items/{index}/additionalClassificationsCount" not in tenders
    assert "/tender/items/additionalClassificationsCount" not in tenders_items


def test_flatten_with_counters(spec, releases):
    releases[0]["tender"]["items"] = releases[0]["tender"]["items"] * 6
    releases[0]["tender"]["items"][0]["additionalClassifications"] = (
        releases[0]["tender"]["items"][0]["additionalClassifications"] * 6
    )
    for _ in spec.process_items(releases):
        pass
    options = FlattenOptions(**{"selection": {"tenders": {"split": True}}, "count": True})
    flattener = Flattener(options, spec.tables)
    for count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            if name == "tenders":
                for row in rows:
                    items = search(f"[{count}].tender.items", releases)
                    if items:
                        assert "/tender/itemsCount" in row
                        assert len(items) == row["/tender/itemsCount"]
            if name == "tenders_items":
                for index, row in enumerate(rows):
                    additional = search(
                        f"[{count}].tender.items[{index}].additionalClassifications",
                        releases,
                    )
                    if additional:
                        assert "/tender/items/additionalClassificationsCount" in row
                        assert len(additional) == row["/tender/items/additionalClassificationsCount"]


def test_flatten_with_repeat(spec_analyzed, releases):
    options = FlattenOptions(
        **{
            "selection": {"tenders": {"split": True, "repeat": ["/tender/id"]}},
        }
    )
    flattener = Flattener(options, spec_analyzed.tables)
    for count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            if name == "tenders":
                continue
            for row in rows:
                assert "id" in row
                assert "ocid" in row
                assert "rowID" in row
                assert "/tender/id" in row
                assert row["/tender/id"] == search(f"[{count}].tender.id", releases)


def test_flatten_with_unnest(spec_analyzed, releases):
    field = "/tender/items/0/id"
    options = FlattenOptions(
        **{
            "selection": {"tenders": {"split": True, "unnest": [field]}},
        }
    )
    flattener = Flattener(options, spec_analyzed.tables)
    for count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            for row in rows:
                if name != "tenders":
                    assert field not in row
                    continue
                item_id = search(f"[{count}].tender.items[0].id", releases)
                if item_id:
                    assert field in row
                    assert search(f"[{count}].tender.items[0].id", releases) == row[field]


def test_flatten_with_exclude(spec, releases):
    releases[0]["tender"]["items"] = releases[0]["tender"]["items"] * 6
    for _ in spec.process_items(releases):
        pass
    options = FlattenOptions(**{"selection": {"tenders": {"split": True}}, "exclude": ["tenders_items"]})
    flattener = Flattener(options, spec.tables)
    all_rows = defaultdict(list)
    for count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            all_rows[name].extend(rows)
    assert "tenders" in all_rows
    assert "tenders_items" not in all_rows

    options = FlattenOptions(**{"selection": {"tenders": {"split": True}}})
    flattener = Flattener(options, spec.tables)
    all_rows = defaultdict(list)
    for count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            all_rows[name].extend(rows)
    assert "tenders" in all_rows
    assert "tenders_items" in all_rows


def test_flatten_with_only(spec_analyzed, releases):
    options = FlattenOptions(
        **{"selection": {"tenders": {"split": True, "only": ["/tender/id"]}, "parties": {"split": False}}}
    )
    flattener = Flattener(options, spec_analyzed.tables)
    all_rows = defaultdict(list)
    for count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            all_rows[name].extend(rows)
    assert all_rows["tenders"]

    for row in all_rows["tenders"]:
        assert not set(row).difference(["/tender/id", "rowID", "ocid", "parentID", "id"])

    options = FlattenOptions(**{"selection": {"tenders": {"split": False, "only": ["/tender/id"]}}})
    flattener = Flattener(options, spec_analyzed.tables)
    all_rows = defaultdict(list)
    for count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            all_rows[name].extend(rows)

    assert all_rows["tenders"]
    for row in all_rows["tenders"]:
        assert not set(row).difference(["/tender/id", "rowID", "ocid", "parentID", "id"])


def test_flatten_should_not_split(spec_analyzed, releases):
    options = FlattenOptions(**{"selection": {"tenders": {"split": False}}})
    flattener = Flattener(options, spec_analyzed.tables)
    all_rows = defaultdict(list)
    for count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            all_rows[name].extend(rows)
    assert "tender_items" not in all_rows
    assert "tenders_items_addit" not in all_rows
    tenders = all_rows["tenders"]

    for tender, release in zip(tenders, releases):
        items = release.get("tender", {}).get("items")
        if release.get("tender", {}).get("items"):
            assert "/tender/items/0/id" in tender
            assert "/tender/items/0/description" in tender
            if len(items) > 1:
                assert "/tender/items/1/id" in tender
                assert "/tender/items/1/description" in tender


@pytest.mark.parametrize(
    "options",
    [
        FlattenOptions(**{"selection": {"tenders": {"split": True}, "tenders_items": {"split": False}}}),
        FlattenOptions(**{"selection": {"tenders": {"split": True}}}),
    ],
)
def test_flatten_splitted_with_child(spec, releases, options):
    releases[0]["tender"]["items"] = releases[0]["tender"]["items"] * 6
    for _ in spec.process_items(releases):
        pass
    flattener = Flattener(options, spec.tables)
    all_rows = defaultdict(list)
    for count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            all_rows[name].extend(rows)

    assert "tenders_items" in all_rows
    assert "tenders_items_addit" not in all_rows
    tenders = all_rows["tenders"]

    for tender, release in zip(tenders, releases):
        if release.get("tender", {}).get("items"):
            assert "/tender/items/0/id" not in tender
            assert "/tender/items/0/description" not in tender
            assert "/tender/items/1/id" not in tender
            assert "/tender/items/1/description" not in tender

    items = all_rows["tenders_items"]
    for item in items:
        assert "/tender/items/id" in item
        assert "/tender/items/description" in item


def test_flatten_string_arrays(spec_analyzed, releases):
    options = FlattenOptions(
        **{
            "selection": {"tenders": {"split": True}, "parties": {"split": True}},
            "exclude": ["tenders_items", "parties_ids", "tenders_tenderers"],
        }
    )
    flattener = Flattener(options, spec_analyzed.tables)
    fields = ["submissionMethod", "roles"]
    for count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            counters = defaultdict(int)
            for row in reversed(rows):
                for key, value in row.items():
                    if "/" in key:
                        key = key.replace("parties", f"parties/{counters['parties']}")
                        actual = resolve_pointer(releases[count], key)
                        if any(key.endswith(field) for field in fields):
                            actual = JOINABLE_SEPARATOR.join(actual)
                        assert actual == value
                counters[name] += 1


def test_flatten_only_no_default_columns(spec_analyzed, releases):
    options = FlattenOptions(**{"selection": {"tenders": {"split": False, "only": ["/tender/id"]}}})
    flattener = Flattener(options, spec_analyzed.tables)
    for _count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            for row in rows:
                assert not set(row.keys()).difference(set(["/tender/id"]))


def test_flatten_buyer(spec_analyzed, releases):
    options = FlattenOptions(**{"selection": {"parties": {"split": True}}, "exclude": ["parties_ids"]})
    flattener = Flattener(options, spec_analyzed.tables)
    for count, flat in flattener.flatten(releases):
        buyer = search(f"[{count}].buyer", releases)
        for name, rows in flat.items():
            for row in rows:
                if buyer:
                    assert "/buyer/id" in row
                    assert "/buyer/name" in row



from collections import defaultdict

from spoonbill.utils import SchemaHeaderExtractor


class BaseWriter:
    """
    Base writer class
    """

    def __init__(self, workdir, tables, options, schema):
        """
        :param workdir: Working directory
        :param tables: The table objects
        :param options: Flattening options
        """
        self.workdir = workdir
        self.tables = tables
        self.options = options
        self.names = {}
        self.headers = {}
        self.names_counter = defaultdict(int)
        self.schema = schema
        self.schema_headers = SchemaHeaderExtractor(self.schema)

    def get_headers(self, table, options):
        """
        Return a table's headers, respecting the human and override options.

        :param table: A table object
        :param options: Flattening options
        :return: Mapping between the machine-readable headers and the output headers
        """
        split = options.split and (table.rolled_up or table.splitted)
        # split = options.split and table.splitted
        if table.parent != "" and table.child_tables:
            split = True
        headers = {c: c for c in table.available_rows(split=split)}
        if (
            table.parent != ""
            and table.parent.name in self.options.selection  # noqa: W503
            and self.options.selection[table.parent.name].pretty_headers  # noqa:W503
        ):
            self.options.selection[table.name].pretty_headers = True
        if options.pretty_headers:
            for c in headers:
                headers[c] = table.titles.get(c, c)
                for k, v in headers.items():
                    headers[k] = self.schema_headers.get_header(k, v)
            for col in headers.keys():
                for char in col:
                    if char.isnumeric() and char != "0":
                        title_col = col.replace(char, "0")
                        headers[col] = headers[title_col]

        if options.headers:
            for c, h in options.headers.items():
                headers[c] = h
        return headers

    def _name_check(self, table_name):
        self.names_counter[table_name] += 1
        if self.names_counter[table_name] > 1:
            table_name += str(self.names_counter[table_name] - 1)
        return table_name

    def init_sheet(self, name, table):
        """
        Initialize a sheet, setting its headers and unique name.

        In this context, the sheet might be either a CSV file or a sheet in an Excel workbook.

        :param name: Table name
        :param table: Table object
        """
        options = self.options.selection[name]
        self.headers[name] = self.get_headers(table, options)
        self.names[name] = self._name_check(options.name or name)
        return self.names[name], self.headers[name]



SEPARATOR = "/"

# we can try to infer tables from  schema
# but it may require some heuristics or handling exceptional cases
# like "tenders" which is not array and called "tender"
# so this way seems like middle ground between flexibility and simplicity
ROOT_TABLES = {
    "tenders": ["/tender"],
    "awards": ["/awards"],
    "contracts": ["/contracts"],
    "planning": ["/planning"],
    "parties": ["/parties", "/buyer"],
}
# TODO: extract types from schema? items and additionalClassifications as table?
COMBINED_TABLES = {
    "documents": [
        "/planning/documents",
        "/tender/documents",
        "/awards/documents",
        "/contracts/documents",
        "/contracts/implementation/documents",
    ],
    "milestones": [
        "/planning/milestones",
        "/tender/milestones",
        "/contracts/milestones",
        "/contracts/implementation/milestones",
    ],
    "amendments": [
        "/planning/amendments",
        "/tender/amendments",
        "/awards/amendments",
        "/contracts/amendments",
        "/contracts/implementation/amendments",
    ],
}

PREVIEW_ROWS = 5


DEFAULT_FIELDS = ["ocid", "id", "rowID", "parentID"]
DEFAULT_FIELDS_COMBINED = ["ocid", "id", "rowID", "parentID", "parentTable"]

ARRAY = "array of {}"
# TODO: is joinable good name?
JOINABLE = "joinable"
JOINABLE_SEPARATOR = ";"
TABLE_THRESHOLD = 5
CURRENT_SCHEMA_TAG = "1__1__5"
CURRENT_URL_TAG = "1.1"
DEFAULT_SCHEMA_URL = {
    "releases": {
        "en": f"https://standard.open-contracting.org/{CURRENT_URL_TAG}/en/release-package-schema.json",
        "es": f"https://standard.open-contracting.org/{CURRENT_URL_TAG}/es/release-package-schema.json",
    },
    "records": {
        "en": f"https://standard.open-contracting.org/{CURRENT_URL_TAG}/en/record-package-schema.json",
        "es": f"https://standard.open-contracting.org/{CURRENT_URL_TAG}/es/record-package-schema.json",
    },
}



import logging
from collections import defaultdict, deque
from dataclasses import dataclass, field, is_dataclass, replace
from typing import List, Mapping

from spoonbill.common import DEFAULT_FIELDS, JOINABLE, JOINABLE_SEPARATOR, SEPARATOR
from spoonbill.i18n import LOCALE, _
from spoonbill.rowdata import Rows
from spoonbill.spec import Table
from spoonbill.utils import get_pointer, get_root, make_count_column

LOGGER = logging.getLogger("spoonbill")


@dataclass
class TableFlattenConfig:
    """Table specific flattening configuration

    :param split: Split child arrays to separate tables
    :param pretty_headers: Use human friendly headers extracted from schema
    :param headers: User edited headers to override automatically extracted
    :param unnest: List of columns to output from child to parent table
    :param repeat: List of columns to clone in child tables
    :param only: List of columns to output
    :param name: Overwrite table name
    """

    split: bool
    pretty_headers: bool = False
    headers: Mapping[str, str] = field(default_factory=dict)
    repeat: List[str] = field(default_factory=list)
    unnest: List[str] = field(default_factory=list)
    only: List[str] = field(default_factory=list)
    name: str = ""


@dataclass
class FlattenOptions:
    """Flattening configuration

    :param selection: List of selected tables to extract from data
    :param count: Include number of rows in child table in each parent table
    :param exclude: List of tables to exclude from export
    """

    selection: Mapping[str, TableFlattenConfig]
    exclude: List[str] = field(default_factory=list)
    count: bool = False

    def __post_init__(self):
        for name, table in self.selection.items():
            if not is_dataclass(table):
                self.selection[name] = TableFlattenConfig(**table)


class Flattener:
    """Data flattener

    In order to export data correctly Flattener requires previously analyzed tables data.
    During the process flattener could add columns not based on schema analysis, such as
    `itemsCount`.
    In every generated row, depending on table type, flattener will always few add autogenerated columns.
    For root table:
    * rowID
    * id
    * ocid

    For child tables this list well be extended with `parentID` column.

    :param options: Flattening options
    :param tables: Analyzed tables data
    :param language: Language to use for the human-readable headings
    """

    def __init__(self, options: FlattenOptions, tables: Mapping[str, Table], language=LOCALE):
        if not is_dataclass(options):
            options = FlattenOptions(**options)
        self.options = options
        self.tables = tables
        self.language = language

        self._lookup_map = {}
        self._types_map = {}
        self._path_map = {}
        self.init_table_selection(tables)
        self.init_options(self.tables)

    def init_table_selection(self, tables):
        self.tables = {}
        for name, table in tables.items():
            if name not in self.options.selection:
                continue
            options = self.options.selection[name]
            if options.only:
                self.init_only(table, options.only, options.split)
            self.init_table_lookup(self.tables, table)
            self.init_child_tables(tables, table, options)

    def init_child_tables(self, tables, table, options):
        for c_name in table.child_tables:
            if c_name in self.options.exclude:
                continue
            c_table = tables[c_name]
            if c_table.rolled_up or options.split:
                if c_table.name not in self.options.selection:
                    self.options.selection[c_table.name] = replace(options)
                self.init_table_lookup(self.tables, c_table)
            else:
                # use parent table
                target = table
                while target.parent and not target.rolled_up:
                    target = target.parent
                self.init_table_lookup(self.tables, target)
            self.init_child_tables(tables, c_table, options)

    def init_map(self, map, paths, table, only=None, target=None):
        if not target:
            target = table
        for path in paths:
            if path not in DEFAULT_FIELDS:
                if not only or (only and path in only):
                    map[path] = target

    def _map_path(self, table, target=None):
        self.init_map(self._path_map, table.path, table, target=target)

    def _map_types(self, table, target=None):
        self.init_map(self._types_map, table.types, table, target=target)

    def _map_cols(self, table, split, target=None):
        cols = table if split else table.combined_columns
        self.init_map(self._lookup_map, cols, table, target=target)

    def init_table_lookup(self, tables, table, target=None):
        if table.total_rows == 0:
            return
        name = table.name
        tables[name] = table
        self._map_path(table, target=target)
        self._map_types(table, target=target)
        self._map_cols(table, self.options.selection[name].split, target=target)

    def init_count(self, table, options):
        if not table.splitted:
            return
        for array in table.arrays:
            path = make_count_column(array)
            # add count columns only if table is rolled up
            # in other way it could be frustrating
            # e.g. it may generate columns for whole array like:
            # /tender/items/200/additionalClassificationsCount
            table.add_column(path, "integer", path, additional=True, propagated=False, header=path)
            table.inc_column(path, path)

    def init_unnest(self, table, options):
        for col_id in options.unnest:
            col = table.combined_columns.get(col_id)
            if col:
                table.columns[col_id] = col

    def init_repeat(self, table, options):
        for col_id in options.repeat:
            columns = table.columns if options.split else table.combined_columns
            col = columns.get(col_id)
            if not col:
                LOGGER.warning(_("Ignoring repeat column {} because it is not in table {}").format(col_id, table.name))
                continue
            for c_name in table.child_tables:
                child_table = self.tables.get(c_name)
                if child_table:
                    child_table.columns[col_id] = col
                    child_table.combined_columns[col_id] = col
                    child_table.titles[col_id] = col_id

    def init_options(self, tables):
        for table in tables.values():
            options = self.options.selection[table.name]

            if self.options.count:
                self.init_count(table, options)
            if options.unnest:
                self.init_unnest(table, options)
            if options.repeat:
                self.init_repeat(table, options)

    def init_only(self, table, only, split):
        paths = {c_id: c for c_id, c in table.types.items() if c_id not in table.combined_columns}
        table.filter_columns(lambda col: col.id not in only)
        paths.update({c_id: c.type for c_id, c in table.columns.items()})
        table.types = paths

    def get_table(self, pointer):
        return self._lookup_map.get(pointer) or self._types_map.get(pointer)

    def flatten(self, releases):
        """Flatten releases

        :param releases: releases as iterable object
        :return: Iterator over mapping between table name and list of rows for each release
        """

        for counter, release in enumerate(releases):
            to_flatten = deque([("", "", "", {}, release, {})])
            rows = Rows(ocid=release["ocid"], buyer=release.get("buyer", {}), data=defaultdict(list))

            while to_flatten:
                abs_path, path, parent_key, parent, record, repeat = to_flatten.pop()

                table = self._path_map.get(path)
                if path == "/buyer":
                    # only useful in analysis
                    continue
                if table:
                    # Strict match /tender /parties etc., so this is a new row
                    row = rows.new_row(table, record.get("id", ""))
                    only = self.options.selection[table.name].only
                    if only:
                        row = {col: col_v for col, col_v in row.items() if col in only}
                    if table.is_root:
                        repeat = {}
                    if repeat:
                        row.update(repeat)
                    rows.data[table.name].append(row)
                for key, item in record.items():
                    pointer = SEPARATOR.join((path, key))
                    abs_pointer = SEPARATOR.join((abs_path, key))
                    table = self.get_table(pointer)
                    if not table:
                        continue

                    item_type = table.types.get(pointer)
                    options = self.options.selection[table.name]
                    split = options.split
                    if pointer in options.repeat:
                        repeat[pointer] = item

                    if isinstance(item, dict):
                        to_flatten.append((abs_pointer, pointer, key, record, item, repeat))
                    elif isinstance(item, list):
                        if item_type == JOINABLE:
                            value = JOINABLE_SEPARATOR.join((str(i) for i in item))
                            rows.data[table.name][-1][pointer] = value
                        else:
                            if self.options.count and table.splitted:
                                abs_pointer = get_pointer(
                                    table,
                                    abs_pointer,
                                    pointer,
                                    split,
                                )
                                abs_pointer += "Count"
                                if abs_pointer in table:
                                    rows.data[table.name][-1][abs_pointer] = len(item)
                            for index, value in enumerate(item):
                                if isinstance(value, dict):
                                    abs_pointer = get_pointer(
                                        table,
                                        SEPARATOR.join((abs_path, key)),
                                        pointer,
                                        split,
                                        index=str(index),
                                    )
                                    to_flatten.append(
                                        (
                                            abs_pointer,
                                            pointer,
                                            key,
                                            record,
                                            value,
                                            repeat,
                                        )
                                    )
                    else:
                        if table.is_combined:
                            pointer = SEPARATOR + SEPARATOR.join((parent_key, key))
                            abs_pointer = pointer
                        if not table.is_root:
                            root = get_root(table)
                            unnest = self.options.selection[root.name].unnest
                            if unnest and abs_pointer in unnest:
                                rows.data[root.name][-1][abs_pointer] = item
                                continue

                        pointer = get_pointer(table, abs_pointer, pointer, split)
                        if pointer in table.combined_columns:
                            rows.data[table.name][-1][pointer] = item
            yield counter, rows



import csv
import logging

from spoonbill.i18n import _
from spoonbill.writers.base_writer import BaseWriter

LOGGER = logging.getLogger("spoonbill")


class CSVWriter(BaseWriter):
    """
    Writer class with output to CSV files.

    For each table, a corresponding CSV file will be created.
    """

    name = "csv"

    def __init__(self, workdir, tables, options, schema):
        """
        :param workdir: Working directory
        :param tables: The table objects
        :param options: Flattening options
        """

        super().__init__(workdir, tables, options, schema=schema)
        self.writers = {}
        self.fds = []

    def __enter__(self):
        """
        Write the headers to the output file.
        """

        for name, table in self.tables.items():
            table_name, headers = self.init_sheet(name, table)

            try:
                path = self.workdir / f"{table_name}.csv"
                LOGGER.info(_("Dumping table '{}' to file '{}'").format(table_name, path))
                fd = open(path, "w")
            except (IOError, OSError) as e:
                LOGGER.error(_("Failed to open file {} with error {}").format(path, e))
                return
            writer = csv.DictWriter(fd, headers)
            self.fds.append(fd)
            self.writers[name] = writer

        for name, writer in self.writers.items():
            headers = self.headers[name]
            try:
                writer.writerow(headers)
            except ValueError as err:
                LOGGER.error(_("Failed to headers with error {}").format(err))
        return self

    def __exit__(self, *args):
        """
        Close the CSV files.
        """

        for fd in self.fds:
            fd.close()

    def writerow(self, table, row):
        """
        Write a row to the output file.
        """

        try:
            self.writers[table].writerow(row)
        except ValueError as err:
            LOGGER.error(_("Operation produced invalid path. This a software bug, please send issue to developers"))
            LOGGER.error(_("Failed to write row {} with error {}").format(row.get("rowID"), err))
        except KeyError:
            LOGGER.error(_("Invalid table {}").format(table))



import logging
from operator import attrgetter
from pathlib import Path

import jsonref
from ocdsextensionregistry import ProfileBuilder
from ocdskit.util import detect_format

from spoonbill.common import COMBINED_TABLES, CURRENT_SCHEMA_TAG, DEFAULT_SCHEMA_URL, ROOT_TABLES, TABLE_THRESHOLD
from spoonbill.flatten import Flattener
from spoonbill.i18n import LOCALE, _
from spoonbill.stats import DataPreprocessor
from spoonbill.utils import get_order, get_reader, iter_file, resolve_file_uri
from spoonbill.writers import CSVWriter, XlsxWriter

LOGGER = logging.getLogger("spoonbill")


class FileAnalyzer:
    """Main utility for analyzing files
    :param workdir: Working directory
    :param schema: Json schema file to use with data
    :param root_tables: Path configuration which should become root tables
    :param combined_tables: Path configuration for tables with multiple sources
    :param pkg_type: Field name to access records
    :param language: Language to use for the human-readable headings
    :param table_threshold: The maximum number of elements in an array before it is split into a table
    """

    def __init__(
        self,
        workdir,
        schema=None,
        state_file=None,
        root_tables=ROOT_TABLES,
        combined_tables=COMBINED_TABLES,
        pkg_type="releases",
        language=LOCALE,
        table_threshold=TABLE_THRESHOLD,
    ):
        self.workdir = Path(workdir)
        self.multiple_values = False
        self.schema = schema
        self.root_tables = root_tables
        self.combined_tables = combined_tables
        self.language = language
        self.table_threshold = table_threshold
        if state_file:
            self.spec = DataPreprocessor.restore(state_file)
            self.sort_tables()
        else:
            self.spec = None
        self.pkg_type = pkg_type
        self.order = None

    def analyze_file(self, filenames, with_preview=True):

        """Analyze provided file
        :param filename: Input filename
        :param with_preview: Generate preview during analysis
        """
        if not isinstance(filenames, list):
            filenames = [filenames]
        path = self.workdir / filenames[0]
        (
            input_format,
            _is_concatenated,
            _is_array,
        ) = detect_format(path=path, reader=get_reader(path))
        LOGGER.info(_("Input file is {}").format(input_format))
        self.multiple_values = _is_concatenated
        self.parse_schema(input_format, self.schema)
        if self.spec is None:
            self.spec = DataPreprocessor(
                self.schema,
                self.root_tables,
                combined_tables=self.combined_tables,
                language=self.language,
                table_threshold=self.table_threshold,
                multiple_values=self.multiple_values,
                pkg_type=self.pkg_type,
            )
        for filename in filenames:
            path = self.workdir / filename
            reader = get_reader(path)
            with reader(path, "rb") as fd:
                items = iter_file(fd, self.pkg_type, multiple_values=self.multiple_values)
                for count in self.spec.process_items(items):
                    yield fd.tell(), count
        self.sort_tables()

    def dump_to_file(self, filename):
        """Save analyzed information to file
        :param filename: Output filename in working directory
        """
        path = self.workdir / filename
        self.spec.dump(path)

    def parse_schema(self, input_format, schema=None):
        if schema:
            schema = resolve_file_uri(schema)
        if "release" in input_format:
            pkg_type = "releases"
            getter = attrgetter("release_package_schema")
        else:
            pkg_type = "records"
            getter = attrgetter("record_package_schema")
        url = DEFAULT_SCHEMA_URL[pkg_type].get(self.language[:2], DEFAULT_SCHEMA_URL[pkg_type]["en"])
        if not schema:
            LOGGER.info(_("No schema provided, using version {}").format(CURRENT_SCHEMA_TAG))
            profile = ProfileBuilder(CURRENT_SCHEMA_TAG, {}, schema_base_url=url)
            schema = getter(profile)()
        title = schema.get("title", "").lower()
        if not title:
            raise ValueError(_("Incomplete schema, please make sure your data is correct"))
        if "package" in title:
            # TODO: is is a good way to get release/record schema
            schema = jsonref.JsonRef.replace_refs(schema)
            schema = schema["properties"][pkg_type]["items"]

        self.schema = schema
        self.pkg_type = pkg_type

    def sort_tables(self):
        """
        Sort tables according to order of arrays in schema
        :return:
        """
        self.order = get_order(self.spec.schema["properties"].keys())
        out_schema_tables = {
            name: table for name, table in self.spec.tables.items() if name.split("_")[0] not in self.order
        }
        within_schema_tables = {
            name: table for name, table in self.spec.tables.items() if name.split("_")[0] in self.order
        }

        sorted_tables = dict(
            sorted(
                within_schema_tables.items(),
                key=lambda sheet: self.order.index(sheet[0].split("_")[0])
                if sheet[0].split("_")[0] in self.order
                else -1,
            )
        )
        self.spec.tables = {**sorted_tables, **out_schema_tables}


class FileFlattener:
    """Main utility for flattening files
    :param workdir: Working directory
    :param options: Flattening configuration
    :param analyzer: Analyzed data object
    :param pkg_type: Field name to access records
    :param csv: If True generate cvs files
    :param xlsx: Generate combined xlsx table
    :param language: Language to use for the human-readable headings
    """

    def __init__(
        self,
        workdir,
        options,
        analyzer=None,
        tables=None,
        pkg_type="releases",
        csv=None,
        xlsx="result.xlsx",
        language=LOCALE,
        multiple_values=False,
        schema=None,
    ):
        self.tables = tables if tables else analyzer.spec.tables
        self.flattener = Flattener(options, self.tables, language=language)
        self.workdir = Path(workdir)
        # TODO: detect package, where?
        self.writers = []
        self.csv = csv
        self.xlsx = xlsx
        self.multiple_values = multiple_values if multiple_values else analyzer.multiple_values if analyzer else False
        self.pkg_type = pkg_type if pkg_type else analyzer.pkg_type if analyzer else "releases"
        self.schema = schema or analyzer.spec.schema

    def _flatten(self, filenames, writers):
        if not isinstance(filenames, list):
            filenames = [filenames]
        for filename in filenames:
            path = self.workdir / filename
            reader = get_reader(path)
            with reader(path, "rb") as fd:
                items = iter_file(fd, self.pkg_type, multiple_values=self.multiple_values)
                for count, data in self.flattener.flatten(items):
                    for table, rows in data.items():
                        for row in rows:
                            for wr in writers:
                                wr.writerow(table, row)
                    yield count

    def flatten_file(self, filename):
        """Flatten file
        :param filename: Input filename in working directory
        """
        workdir = self.workdir

        if isinstance(self.csv, Path):
            workdir = self.csv
        if not self.xlsx and self.csv:
            with CSVWriter(
                workdir,
                self.flattener.tables,
                self.flattener.options,
                schema=self.schema,
            ) as writer:
                for count in self._flatten(filename, [writer]):
                    yield count
        if self.xlsx and not self.csv:
            with XlsxWriter(
                self.workdir,
                self.flattener.tables,
                self.flattener.options,
                filename=self.xlsx,
                schema=self.schema,
            ) as writer:
                for count in self._flatten(filename, [writer]):
                    yield count

        if self.xlsx and self.csv:
            with XlsxWriter(
                self.workdir,
                self.flattener.tables,
                self.flattener.options,
                filename=self.xlsx,
                schema=self.schema,
            ) as xlsx, CSVWriter(
                workdir,
                self.flattener.tables,
                self.flattener.options,
                schema=self.schema,
            ) as csv:
                for count in self._flatten(filename, [xlsx, csv]):
                    yield count


__all__ = ["FileFlattener", "FileAnalyzer"]



from copy import deepcopy

from spoonbill.common import ROOT_TABLES

TEST_ROOT_TABLES = deepcopy(ROOT_TABLES)
TEST_COMBINED_TABLES = {
    "documents": [
        "/planning/documents",
        "/tender/documents",
        "/awards/documents",
        "/contracts/documents",
        "/contracts/implementation/documents",
    ],
    "milestones": [
        "/planning/milestones",
        "/tender/milestones",
        "/contracts/milestones",
        "/contracts/implementation/milestones",
    ],
    "amendments": [
        "/planning/amendments",
        "/tender/amendments",
        "/awards/amendments",
        "/contracts/amendments",
        "/contracts/implementation/amendments",
    ],
}

tenders_columns = [
    "/tender/id",
    "/tender/title",
    "/tender/description",
    "/tender/status",
    "/tender/procurementMethod",
    "/tender/procurementMethodDetails",
    "/tender/procurementMethodRationale",
    "/tender/mainProcurementCategory",
    "/tender/additionalProcurementCategories",
    "/tender/awardCriteria",
    "/tender/awardCriteriaDetails",
    "/tender/submissionMethodDetails",
    "/tender/submissionMethod",
    "/tender/hasEnquiries",
    "/tender/eligibilityCriteria",
    "/tender/numberOfTenderers",
    "/tender/contractPeriod/startDate",
    "/tender/contractPeriod/endDate",
    "/tender/contractPeriod/maxExtentDate",
    "/tender/contractPeriod/durationInDays",
    "/tender/awardPeriod/startDate",
    "/tender/awardPeriod/endDate",
    "/tender/awardPeriod/maxExtentDate",
    "/tender/awardPeriod/durationInDays",
    "/tender/enquiryPeriod/startDate",
    "/tender/enquiryPeriod/endDate",
    "/tender/enquiryPeriod/maxExtentDate",
    "/tender/enquiryPeriod/durationInDays",
    "/tender/tenderPeriod/startDate",
    "/tender/tenderPeriod/endDate",
    "/tender/tenderPeriod/maxExtentDate",
    "/tender/tenderPeriod/durationInDays",
    "/tender/minValue/amount",
    "/tender/minValue/currency",
    "/tender/value/amount",
    "/tender/value/currency",
    "/tender/procuringEntity/name",
    "/tender/procuringEntity/id",
]
tenders_arrays = [
    "/tender/items",
    "/tender/tenderers",
    "/tender/items/additionalClassifications",
]
tenders_combined_columns = [
    "/tender/id",
    "/tender/title",
    "/tender/description",
    "/tender/status",
    "/tender/procurementMethod",
    "/tender/procurementMethodDetails",
    "/tender/procurementMethodRationale",
    "/tender/mainProcurementCategory",
    "/tender/additionalProcurementCategories",
    "/tender/awardCriteria",
    "/tender/awardCriteriaDetails",
    "/tender/submissionMethod",
    "/tender/submissionMethodDetails",
    "/tender/hasEnquiries",
    "/tender/eligibilityCriteria",
    "/tender/numberOfTenderers",
    "/tender/tenderers/0/name",
    "/tender/tenderers/0/id",
    "/tender/contractPeriod/startDate",
    "/tender/contractPeriod/endDate",
    "/tender/contractPeriod/maxExtentDate",
    "/tender/contractPeriod/durationInDays",
    "/tender/awardPeriod/startDate",
    "/tender/awardPeriod/endDate",
    "/tender/awardPeriod/maxExtentDate",
    "/tender/awardPeriod/durationInDays",
    "/tender/enquiryPeriod/startDate",
    "/tender/enquiryPeriod/endDate",
    "/tender/enquiryPeriod/maxExtentDate",
    "/tender/enquiryPeriod/durationInDays",
    "/tender/tenderPeriod/startDate",
    "/tender/tenderPeriod/endDate",
    "/tender/tenderPeriod/maxExtentDate",
    "/tender/tenderPeriod/durationInDays",
    "/tender/minValue/amount",
    "/tender/minValue/currency",
    "/tender/value/amount",
    "/tender/value/currency",
    "/tender/items/0/id",
    "/tender/items/0/description",
    "/tender/items/0/quantity",
    "/tender/items/0/unit/scheme",
    "/tender/items/0/unit/id",
    "/tender/items/0/unit/name",
    "/tender/items/0/unit/uri",
    "/tender/items/0/unit/value/amount",
    "/tender/items/0/unit/value/currency",
    "/tender/items/0/additionalClassifications/0/scheme",
    "/tender/items/0/additionalClassifications/0/id",
    "/tender/items/0/additionalClassifications/0/description",
    "/tender/items/0/additionalClassifications/0/uri",
    "/tender/items/0/classification/scheme",
    "/tender/items/0/classification/id",
    "/tender/items/0/classification/description",
    "/tender/items/0/classification/uri",
    "/tender/procuringEntity/name",
    "/tender/procuringEntity/id",
]

awards_columns = [
    "/awards/id",
    "/awards/title",
    "/awards/description",
    "/awards/status",
    "/awards/date",
    "/awards/contractPeriod/startDate",
    "/awards/contractPeriod/endDate",
    "/awards/contractPeriod/maxExtentDate",
    "/awards/contractPeriod/durationInDays",
    "/awards/value/amount",
    "/awards/value/currency",
]
awards_combined_columns = [
    "/awards/id",
    "/awards/title",
    "/awards/description",
    "/awards/status",
    "/awards/date",
    "/awards/contractPeriod/startDate",
    "/awards/contractPeriod/endDate",
    "/awards/contractPeriod/maxExtentDate",
    "/awards/contractPeriod/durationInDays",
    "/awards/items/0/id",
    "/awards/items/0/description",
    "/awards/items/0/quantity",
    "/awards/items/0/unit/scheme",
    "/awards/items/0/unit/id",
    "/awards/items/0/unit/name",
    "/awards/items/0/unit/uri",
    "/awards/items/0/unit/value/amount",
    "/awards/items/0/unit/value/currency",
    "/awards/items/0/additionalClassifications/0/scheme",
    "/awards/items/0/additionalClassifications/0/id",
    "/awards/items/0/additionalClassifications/0/description",
    "/awards/items/0/additionalClassifications/0/uri",
    "/awards/items/0/classification/scheme",
    "/awards/items/0/classification/id",
    "/awards/items/0/classification/description",
    "/awards/items/0/classification/uri",
    "/awards/suppliers/0/name",
    "/awards/suppliers/0/id",
    "/awards/value/amount",
    "/awards/value/currency",
]

awards_arrays = [
    "/awards/suppliers",
    "/awards/items",
    "/awards/items/additionalClassifications",
]
contracts_columns = [
    "/contracts/id",
    "/contracts/awardID",
    "/contracts/title",
    "/contracts/description",
    "/contracts/status",
    "/contracts/dateSigned",
    "/contracts/value/amount",
    "/contracts/value/currency",
    "/contracts/period/startDate",
    "/contracts/period/endDate",
    "/contracts/period/maxExtentDate",
    "/contracts/period/durationInDays",
]
contracts_arrays = [
    "/contracts/items",
    "/contracts/relatedProcesses",
    "/contracts/implementation/transactions",
    "/contracts/items/additionalClassifications",
]
contracts_combined_columns = [
    "/contracts/id",
    "/contracts/awardID",
    "/contracts/title",
    "/contracts/description",
    "/contracts/status",
    "/contracts/dateSigned",
    "/contracts/relatedProcesses/0/id",
    "/contracts/relatedProcesses/0/relationship",
    "/contracts/relatedProcesses/0/title",
    "/contracts/relatedProcesses/0/scheme",
    "/contracts/relatedProcesses/0/identifier",
    "/contracts/relatedProcesses/0/uri",
    "/contracts/implementation/transactions/0/id",
    "/contracts/implementation/transactions/0/source",
    "/contracts/implementation/transactions/0/date",
    "/contracts/implementation/transactions/0/uri",
    "/contracts/implementation/transactions/0/payee/name",
    "/contracts/implementation/transactions/0/payee/id",
    "/contracts/implementation/transactions/0/payer/name",
    "/contracts/implementation/transactions/0/payer/id",
    "/contracts/implementation/transactions/0/value/amount",
    "/contracts/implementation/transactions/0/value/currency",
    "/contracts/items/0/id",
    "/contracts/items/0/description",
    "/contracts/items/0/quantity",
    "/contracts/items/0/unit/scheme",
    "/contracts/items/0/unit/id",
    "/contracts/items/0/unit/name",
    "/contracts/items/0/unit/uri",
    "/contracts/items/0/unit/value/amount",
    "/contracts/items/0/unit/value/currency",
    "/contracts/items/0/additionalClassifications/0/scheme",
    "/contracts/items/0/additionalClassifications/0/id",
    "/contracts/items/0/additionalClassifications/0/description",
    "/contracts/items/0/additionalClassifications/0/uri",
    "/contracts/items/0/classification/scheme",
    "/contracts/items/0/classification/id",
    "/contracts/items/0/classification/description",
    "/contracts/items/0/classification/uri",
    "/contracts/value/amount",
    "/contracts/value/currency",
    "/contracts/period/startDate",
    "/contracts/period/endDate",
    "/contracts/period/maxExtentDate",
    "/contracts/period/durationInDays",
]

planning_columns = [
    "/planning/rationale",
    "/planning/budget/id",
    "/planning/budget/description",
    "/planning/budget/project",
    "/planning/budget/projectID",
    "/planning/budget/uri",
    "/planning/budget/amount/amount",
    "/planning/budget/amount/currency",
]

planning_combined_columns = [
    "/planning/rationale",
    "/planning/budget/id",
    "/planning/budget/description",
    "/planning/budget/project",
    "/planning/budget/projectID",
    "/planning/budget/uri",
    "/planning/budget/amount/amount",
    "/planning/budget/amount/currency",
]
planning_arrays = []
parties_arrays = ["/parties/additionalIdentifiers"]
parties_columns = [
    "/parties/name",
    "/parties/id",
    "/parties/roles",
    "/parties/contactPoint/name",
    "/parties/contactPoint/email",
    "/parties/contactPoint/telephone",
    "/parties/contactPoint/faxNumber",
    "/parties/contactPoint/url",
    "/parties/address/streetAddress",
    "/parties/address/locality",
    "/parties/address/region",
    "/parties/address/postalCode",
    "/parties/address/countryName",
    "/parties/identifier/scheme",
    "/parties/identifier/id",
    "/parties/identifier/legalName",
    "/parties/identifier/uri",
]
parties_combined_columns = [
    "/parties/name",
    "/parties/id",
    "/parties/roles",
    "/parties/roles",
    "/parties/contactPoint/name",
    "/parties/contactPoint/email",
    "/parties/contactPoint/telephone",
    "/parties/contactPoint/faxNumber",
    "/parties/contactPoint/url",
    "/parties/address/streetAddress",
    "/parties/address/locality",
    "/parties/address/region",
    "/parties/address/postalCode",
    "/parties/address/countryName",
    "/parties/additionalIdentifiers/0/scheme",
    "/parties/additionalIdentifiers/0/id",
    "/parties/additionalIdentifiers/0/legalName",
    "/parties/additionalIdentifiers/0/uri",
    "/parties/identifier/scheme",
    "/parties/identifier/id",
    "/parties/identifier/legalName",
    "/parties/identifier/uri",
]



import logging
import pickle
from collections import defaultdict, deque
from functools import lru_cache
from pathlib import Path
from typing import List, Mapping

import jsonref
from flatten_dict import flatten

from spoonbill.common import ARRAY, JOINABLE, JOINABLE_SEPARATOR, PREVIEW_ROWS, SEPARATOR, TABLE_THRESHOLD
from spoonbill.i18n import LOCALE, _
from spoonbill.rowdata import Rows
from spoonbill.spec import Table, add_child_table
from spoonbill.utils import (
    PYTHON_TO_JSON_TYPE,
    RepeatFilter,
    add_paths_to_schema,
    common_prefix,
    extract_type,
    generate_table_name,
    get_matching_tables,
    resolve_file_uri,
    validate_type,
)

LOGGER = logging.getLogger("spoonbill")
LOGGER.addFilter(RepeatFilter())


class DataPreprocessor:
    """
    Data analyzer

    Processes the given schema and, based on this, extracts information from the iterable dataset.

    :param schema: The dataset's schema
    :param root_tables: The paths which should become root tables
    :param combined_tables: The paths which should become tables that combine data from different locations
    :param tables: Use these tables objects instead of parsing the schema
    :param table_threshold: The maximum array length, before it is recommended to split out a child table
    :param total_items: The total objects processed
    :param language: Language to use for the human-readable headings
    """

    def __init__(
        self,
        schema: Mapping,
        root_tables: Mapping[str, List],
        combined_tables: Mapping[str, List] = None,
        tables: Mapping[str, Table] = None,
        table_threshold=TABLE_THRESHOLD,
        total_items=0,
        language=LOCALE,
        multiple_values=False,
        pkg_type=None,
        with_preview=True,
    ):
        self.schema = schema
        self.root_tables = root_tables
        self.combined_tables = combined_tables or {}
        self.tables = tables or {}
        self.table_threshold = table_threshold
        self.multiple_values = multiple_values

        self.total_items = total_items
        self.current_table = None

        self.language = language
        self.names_counter = defaultdict(int)
        self.with_preview = with_preview
        if not self.tables:
            self.parse_schema()
        self.pkg_type = pkg_type

    def __getitem__(self, table):
        return self.tables[table]

    def name_check(self, parent_key, key):
        table_name = generate_table_name(self.current_table.name, parent_key, key)
        self.names_counter[table_name] += 1
        if self.names_counter[table_name] > 1:
            key = key[:4] + str(self.names_counter[table_name] - 1)
        return key

    def guess_type(self, item):
        if isinstance(item, list) and item and not isinstance(item[0], (dict, list)):
            return JOINABLE
        return [PYTHON_TO_JSON_TYPE.get(type(item).__name__)]

    def init_tables(self, tables, is_combined=False):
        """
        Initialize the root tables with default fields.
        """
        for name, path in tables.items():
            table = Table(name, path, is_root=True, is_combined=is_combined, parent="")
            self.tables[name] = table

    def is_base_table(self):
        return self.current_table.is_root or self.current_table.is_combined

    def load_schema(
        self,
    ):
        """"""
        # if isinstance(self.schema, (str, Path)):
        #     self.schema = resolve_file_uri(self.schema)
        # if not isinstance(self.schema, jsonref.JsonRef):
        #     self.schema = jsonref.JsonRef.replace_refs(self.schema)

        if isinstance(self.schema, (str, Path)):
            self.schema = resolve_file_uri(self.schema)
        self.init_tables(self.root_tables)
        if not isinstance(self.schema, jsonref.JsonRef):
            self.schema = jsonref.JsonRef.replace_refs(self.schema)
        if self.combined_tables:
            self.init_tables(self.combined_tables, is_combined=True)

    def prepare_tables(self):
        self.init_tables(self.root_tables)
        if self.combined_tables:
            self.init_tables(self.combined_tables, is_combined=True)

    def parse_schema(self):
        """
        Extract information from the schema.
        """
        self.load_schema()
        # self.prepare_tables()
        proxy = add_paths_to_schema(self.schema)
        to_analyze = deque([("", "", {}, proxy)])

        # TODO: check if recursion is better for field ordering
        while to_analyze:
            path, parent_key, parent, prop = to_analyze.pop()
            if prop.get("deprecated"):
                continue
            # TODO: handle oneOf anyOf allOf
            properties = prop.get("properties", {})
            if properties:
                for key, item in properties.items():
                    if key in ("$title", "$path"):
                        continue
                    if item.get("deprecated"):
                        continue
                    if hasattr(item, "__reference__") and item.__reference__.get("deprecated"):
                        continue

                    typeset = extract_type(item)
                    pointer = self.join_path(path, key)
                    self.current_table = self.get_table(pointer)

                    if not self.current_table:
                        continue

                    self.current_table.types[pointer] = typeset

                    if "object" in typeset:
                        to_analyze.append((pointer, key, properties, item))
                    elif "array" in typeset:
                        items = item["items"]
                        items_type = extract_type(items)
                        if set(items_type) & {"array", "object"}:
                            if pointer not in self.current_table.path:
                                # found child array, need to create child table
                                key = self.name_check(parent_key, key)
                                self._add_table(add_child_table(self.current_table, pointer, parent_key, key), pointer)
                            to_analyze.append((pointer, key, properties, items))
                        else:
                            # This means we in array of strings, so this becomes a single joinable column
                            typeset = ARRAY.format(items_type)
                            self.current_table.types[pointer] = JOINABLE
                            self.current_table.add_column(pointer, typeset, pointer, header=item["$title"])
                    else:
                        if self.current_table.is_combined:
                            pointer = SEPARATOR + self.join_path(parent_key, key)
                        self.current_table.add_column(pointer, typeset, pointer, header=item["$title"])

            else:
                # TODO: not sure what to do here
                continue

    def add_column(self, pointer, typeset):
        self.current_table.add_column(pointer, typeset, pointer)

    def _add_table(self, table, pointer):
        self.tables[table.name] = table
        self.current_table = table
        self.get_table.cache_clear()

    def add_additional_table(self, pointer, abs_pointer, parent_key, key, item):
        LOGGER.debug(_("Detected additional table: %s") % pointer)
        self.current_table.types[pointer] = ["array"]
        self._add_table(add_child_table(self.current_table, pointer, parent_key, key), pointer)
        # add columns beforehand because it might be required
        # to recalculate  and reorder headers when enlarging array
        # there must be a better way but it should work for now
        for extended_item in item:
            for path_, it in flatten(extended_item, reducer="path").items():
                ppointer = self.join_path(pointer, path_)
                if ppointer not in self.current_table:
                    self.current_table.add_column(
                        ppointer,
                        self.guess_type(it),
                        ppointer,
                        abs_path=self.join_path(abs_pointer, path_),
                        header=ppointer,
                    )

    @lru_cache(maxsize=None)
    def get_table(self, path):
        """
        Get the table that best matches the given path.

        :param path: A path
        :return: A table
        """
        candidates = get_matching_tables(self.tables, path)
        if not candidates:
            return
        return candidates[0]

    def add_preview_row(self, rows, item_id, parent_key):
        """
        Append a mostly-empty row to the previews.

        This is important to do, because other code uses an index of -1 to access and update the current row.

        :param rows: The Rows object
        :param item_id: Object id
        """
        table = self.current_table
        if self.with_preview and table.total_rows < PREVIEW_ROWS:
            for p_rows in table.preview_rows, table.preview_rows_combined:
                row = rows.new_row(table, item_id).as_dict()
                p_rows.append(row)

    def inc_table_rows(self, item, rows, parent_key, record):
        c = item if isinstance(item, list) else [item]
        for _nop in c:
            self.current_table.inc()
            self.add_preview_row(rows, record.get("id", ""), parent_key)

    def is_new_row(self, pointer):
        # strict match like /parties, /tender
        return pointer in self.current_table.path and pointer != "/buyer"

    def join_path(self, *args):
        return SEPARATOR.join(args)

    def get_paths_for_combined_table(self, parent_key, key):
        pointer = SEPARATOR + self.join_path(parent_key, key)
        return (pointer, pointer)

    def is_type_matched(self, pointer, item, item_type):
        # TODO: this validation should probably be smarter with arrays
        if item_type and item_type != JOINABLE and not validate_type(item_type, item):
            LOGGER.error("Mismatched type on %s expected %s", pointer, item_type)
            return False
        return True

    def add_joinable_column(self, abs_pointer, pointer):
        LOGGER.debug(_("Detected additional column: %s in %s table") % (abs_pointer, self.current_table.name))
        self.current_table.types[pointer] = JOINABLE
        self.current_table.add_column(
            pointer, JOINABLE, pointer, additional=True, abs_path=abs_pointer, header=pointer
        )

    def handle_array_expanded(self, pointer, item, abs_path, key):
        splitted = len(item) >= self.table_threshold
        if splitted:
            self.current_table.split(pointer)

    def is_array_col(self, abs_path):
        chunks = abs_path.split(SEPARATOR)
        path = self.join_path(*[p for p in chunks if not p.isdigit()])
        return path in self.current_table

    def clean_up_missing_arrays(self):
        def drop(col):
            is_array = table.is_array(col.id)
            return is_array and col.hits == 0

        for table in self.tables.values():
            table.filter_columns(drop)

    def process_items(self, releases, with_preview=True):
        """
        Analyze releases.

        Iterates over every release to calculate metrics and optionally generates previews for combined and split
        versions of each table.

        :param releases: The releases to analyze
        :param with_preview: Whether to generate previews for each table
        """
        for count, release in enumerate(releases):

            to_analyze = deque([("", "", "", {}, release)])
            rows = Rows(ocid=release["ocid"], buyer=release.get("buyer", {}), data=defaultdict(list))
            while to_analyze:
                abs_path, path, parent_key, parent, record = to_analyze.popleft()
                if hasattr(record, "items"):
                    for key, item in record.items():
                        pointer = self.join_path(path, key)

                        self.current_table = self.get_table(pointer)
                        if not self.current_table:
                            continue

                        if self.is_new_row(pointer):
                            self.inc_table_rows(item, rows, parent_key, record)

                        self.extend_table_types(pointer, item)
                        item_type = self.current_table.types.get(pointer)

                        if not self.is_type_matched(pointer, item, item_type):
                            continue

                        if isinstance(item, dict):
                            to_analyze.append(
                                (
                                    self.join_path(abs_path, key),
                                    pointer,
                                    key,
                                    record,
                                    item,
                                )
                            )
                        elif item and isinstance(item, list):
                            abs_pointer = self.join_path(abs_path, key)

                            if not isinstance(item[0], dict) and not item_type:
                                item_type = JOINABLE
                                self.add_joinable_column(abs_pointer, pointer)

                            if item_type == JOINABLE:
                                if pointer not in self.current_table:
                                    self.add_joinable_column(abs_pointer, pointer)
                                self.current_table.inc_column(abs_pointer, pointer)
                                if self.with_preview and count < PREVIEW_ROWS:
                                    value = JOINABLE_SEPARATOR.join([str(i) for i in item])
                                    self.current_table.set_preview_path(
                                        abs_pointer, pointer, value, self.table_threshold
                                    )
                            elif self.is_base_table():
                                for value in item:
                                    to_analyze.append(
                                        (
                                            abs_pointer,
                                            pointer,
                                            key,
                                            record,
                                            value,
                                        )
                                    )
                            else:
                                parent_table = self.current_table.parent
                                if pointer not in parent_table.arrays:
                                    LOGGER.debug(_("Detected additional table: %s") % pointer)
                                    self.current_table.types[pointer] = ["array"]
                                    parent_table = self.current_table
                                    self.add_additional_table(pointer, abs_pointer, parent_key, key, item)
                                    self.add_preview_row(rows, record.get("id", ""), parent_key)

                                if parent_table.set_array(pointer, item):
                                    self.handle_array_expanded(pointer, item, abs_path, key)

                                for i, value in enumerate(item):
                                    if isinstance(value, dict):
                                        abs_pointer = self.join_path(abs_path, key, str(i))
                                        to_analyze.append(
                                            (
                                                abs_pointer,
                                                pointer,
                                                parent_key,
                                                record,
                                                value,
                                            )
                                        )
                        else:
                            abs_pointer = self.join_path(abs_path, key)
                            if self.current_table.is_combined:
                                pointer, abs_pointer = self.get_paths_for_combined_table(parent_key, key)
                            col = self.current_table.columns.get(pointer)
                            if col:
                                if abs_pointer not in self.current_table:
                                    parent = self.current_table.parent
                                    parent.add_array_column(col, pointer, abs_pointer, max=self.table_threshold)
                            else:
                                self.current_table.add_column(
                                    pointer,
                                    self.guess_type(item),
                                    pointer,
                                    additional=True,
                                    abs_path=abs_pointer,
                                )
                            self.current_table.inc_column(abs_pointer, pointer)
                            if item and self.with_preview and count < PREVIEW_ROWS:
                                if not pointer.startswith("/buyer"):
                                    self.current_table.set_preview_path(
                                        abs_pointer, pointer, item, self.table_threshold
                                    )
            yield count
        self.clean_up_missing_arrays()
        self.total_items = count

    def dump(self, path):
        """
        Dump the data processor's state to a file.

        :param path: Full path to file
        """
        try:
            with open(path, "wb") as fd:
                pickle.dump(self, fd)
        except (OSError, IOError) as e:
            LOGGER.error(_("Failed to dump DataPreprocessor to file. Error: {}").format(e))

    @classmethod
    def restore(_cls, path):
        """
        Restore a data preprocessor's state from a file.

        :param path: Full path to file
        """
        try:
            with open(path, "rb") as fd:
                return pickle.load(fd)
        except (TypeError, pickle.UnpicklingError):
            LOGGER.error(_("Invalid pickle file. Can't restore."))

    def extend_table_types(self, pointer, item):
        """
        Check if path belong to table and expand its types
        :param pointer: Path to an item
        :param item: Item being analyzed
        """
        table = self.current_table
        if pointer not in table.types and pointer not in table.path:
            if any((common_prefix(pointer, path) for path in table.path)):
                self.current_table.types[pointer] = self.guess_type(item)



import logging
import os
import pathlib
import shutil
from unittest import mock

import openpyxl
from click.testing import CliRunner

from spoonbill.cli import cli
from spoonbill.utils import RepeatFilter

LOGGER = logging.getLogger("spoonbill")
LOGGER.addFilter(RepeatFilter())

FILENAME = pathlib.Path("tests/data/ocds-sample-data.json").absolute()
FILENAME_JSONL = pathlib.Path("tests/data/ocds-sample-data.jsonl").absolute()
FILENAME_GZ = pathlib.Path("tests/data/ocds-sample-data.json.gz").absolute()
EMPTY_LIST_FILE = pathlib.Path("tests/data/empty_list.json").absolute()
SCHEMA = pathlib.Path("tests/data/ocds-simplified-schema.json").absolute()
ANALYZED = pathlib.Path("tests/data/analyzed.state").absolute()
ONLY = pathlib.Path("tests/data/only.txt").absolute()
UNNEST = pathlib.Path("tests/data/unnest.txt").absolute()


def test_no_filename():
    runner = CliRunner()
    result = runner.invoke(cli, [])
    assert result.exit_code == 2
    assert result.output.startswith("Usage")
    assert "Missing argument 'FILENAME'" in result.output


def test_default():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        result = runner.invoke(cli, ["data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output
        assert "Dumping analyzed data" in result.output
        assert "Done flattening. Flattened objects: 6" in result.output


def test_with_selections():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        result = runner.invoke(cli, ["--selection", "tenders", "data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output
        assert "Dumping analyzed data" in result.output
        assert "Done flattening. Flattened objects: 6" in result.output


def test_with_exclude():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        result = runner.invoke(cli, ["--exclude", "tenders_items", "data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output
        assert "Ignoring tables (excluded by user): tenders_items" in result.output
        assert "Dumping analyzed data" in result.output
        assert "Done flattening. Flattened objects: 6" in result.output


def test_with_combine():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        result = runner.invoke(cli, ["--combine", "documents", "data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output
        assert "Dumping analyzed data" in result.output
        assert (
            "Going to export tables: parties,parties_ids,planning,tenders,awards,contracts,documents" in result.output
        )
        assert "Done flattening. Flattened objects: 6" in result.output


def test_table_typo():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        for option in ("--selection", "--combine"):
            result = runner.invoke(cli, [option, "missing", "data.json"])
            assert result.exit_code == 2
            assert "Invalid value" in result.output
            assert "Wrong selection, table 'missing' does not exist"


def test_with_schema():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        shutil.copyfile(SCHEMA, "schema.json")
        result = runner.invoke(cli, ["--selection", "tenders", "--schema", "schema.json", "data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output
        assert "Dumping analyzed data" in result.output
        assert "Done flattening. Flattened objects: 6" in result.output
        assert "Going to export tables: tenders" in result.output


def test_state_file():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        shutil.copyfile(ANALYZED, "analyzed.json")
        result = runner.invoke(cli, ["--state-file", "analyzed.json", "data.json"])
        assert result.exit_code == 0
        assert "Restoring from provided state file" in result.output
        assert (
            "Going to export tables: parties,planning,tenders,awards,contracts,documents,milestones,amendments"
            in result.output
        )
        assert "Done flattening. Flattened objects: 6" in result.output


def test_only():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        shutil.copyfile(ANALYZED, "analyzed.json")
        result = runner.invoke(cli, ["--only", "/tender/title", "data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output
        assert "Using only columns" in result.output
        assert "Done flattening. Flattened objects: 6" in result.output


def test_only_file():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        shutil.copyfile(ANALYZED, "analyzed.json")
        shutil.copyfile(ONLY, "only")
        result = runner.invoke(cli, ["--only-file", "only", "data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output
        assert "Using only columns" in result.output
        assert "Done flattening. Flattened objects: 6" in result.output


def test_repleat():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        shutil.copyfile(ANALYZED, "analyzed.json")
        result = runner.invoke(cli, ["--repeat", "/tender/id", "data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output
        assert "Repeating columns /tender/id in all child table of tenders" in result.output
        assert "Done flattening. Flattened objects: 6" in result.output


def test_repeat_file():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        shutil.copyfile(ANALYZED, "analyzed.json")
        shutil.copyfile(ONLY, "only")
        result = runner.invoke(cli, ["--repeat-file", "only", "data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output
        assert "Repeating columns /tender/id,/tender/title in all child table of tenders" in result.output
        assert "Done flattening. Flattened objects: 6" in result.output


def test_unnest():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        shutil.copyfile(ANALYZED, "analyzed.json")
        result = runner.invoke(cli, ["--unnest", "/tender/items/0/id", "data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output
        assert "Unnesting columns /tender/items/0/id for table tenders" in result.output
        assert "Done flattening. Flattened objects: 6" in result.output


def test_unnest_file():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        shutil.copyfile(ANALYZED, "analyzed.json")
        shutil.copyfile(UNNEST, "unnest")
        result = runner.invoke(cli, ["--unnest-file", "unnest", "data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output
        assert "Unnesting columns /tender/items/0/id for table tenders" in result.output
        assert "Done flattening. Flattened objects: 6" in result.output


def test_table_stats():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        result = runner.invoke(cli, ["data.json"])
        for msg in (
            "Analyze options:",
            " - threshold                      => 5",
            " - language                       => en",
        ):
            assert msg in result.output
        for msg in (
            "Processed tables:",
            " - tenders                        => 4 rows",
            " - awards                         => 3 rows",
            " - contracts                      => 2 rows",
            " - planning                       => 1 rows",
            " - parties                        => 8 rows",
            " ---- parties_ids                 => 14 rows",
        ):
            assert msg in result.output


def test_message_repeat(capsys):
    message = "Around the world, around the world"

    LOGGER.warning(message)
    LOGGER.warning(message)
    LOGGER.warning(message)
    captured = capsys.readouterr()
    assert captured.out.count(message) == 1


def test_xlsx():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        shutil.copyfile(ANALYZED, "analyzed.json")
        result = runner.invoke(cli, ["--state-file", "analyzed.json", "--xlsx", "test.xlsx", "data.json"])
        assert result.exit_code == 0
        assert "Done flattening. Flattened objects: 6" in result.output
        path = pathlib.Path("test.xlsx")
        assert path.resolve().exists()


def test_csv():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        shutil.copyfile(ANALYZED, "analyzed.json")
        os.mkdir("test")
        result = runner.invoke(cli, ["--state-file", "analyzed.json", "--csv", "test", "data.json"])
        assert result.exit_code == 0
        assert "Done flattening. Flattened objects: 6" in result.output
        path = pathlib.Path("test").resolve() / "tenders.csv"
        assert path.resolve().exists()


def test_empty_list_file():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(EMPTY_LIST_FILE, "data.json")
        result = runner.invoke(cli, ["data.json"])
        assert result.exit_code == 0


def test_default_field_only():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        result = runner.invoke(cli, ["--selection", "tenders,parties", "--only", "parentID", "data.json"])
        assert "Using only columns parentID for table tenders" in result.output
        assert "Using only columns parentID for table parties" in result.output


def test_jsonl():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME_JSONL, "data.json")
        result = runner.invoke(cli, ["data.json"])
        assert "Input file is release" in result.output
        assert result.exit_code == 0


def test_gz():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME_GZ, "data.json.gz")
        result = runner.invoke(cli, ["data.json.gz"])
        assert result.exit_code == 0


def test_sheet_order():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        result = runner.invoke(cli, ["data.json"])
        assert result.exit_code == 0
        path = pathlib.Path("result.xlsx")
        assert path.resolve().exists()
        workbook = openpyxl.load_workbook(path)
        sheets = [name for name in workbook.sheetnames]
        assert sheets == [
            "parties",
            "parties_ids",
            "planning",
            "tenders",
            "awards",
            "contracts",
            "documents",
            "milestones",
            "amendments",
        ]


def test_sheet_order_state_file():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        shutil.copyfile(ANALYZED, "analyzed.json")
        result = runner.invoke(cli, ["--state-file", "analyzed.json", "data.json"])
        assert result.exit_code == 0
        path = pathlib.Path("result.xlsx")
        assert path.resolve().exists()
        workbook = openpyxl.load_workbook(path)
        sheets = [name for name in workbook.sheetnames]
        assert sheets == [
            "parties",
            "planning",
            "tenders",
            "awards",
            "contracts",
            "documents",
            "milestones",
            "amendments",
        ]


@mock.patch.dict(os.environ, {"LANG": "uk_UA"})
def test_locale_not_found():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        result = runner.invoke(cli, ["data.json"])
        assert result.exit_code == 0
        assert "Input file is release package" in result.output


def test_pretty_headers_en():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        result = runner.invoke(cli, ["--human", "--language", "en", "data.json"])
        assert result.exit_code == 0
        path = pathlib.Path("result.xlsx")
        assert path.resolve().exists()
        xlsx_reader = openpyxl.load_workbook(path)
        sheet = xlsx_reader["parties"]
        header_values = [cell.value for cell in sheet[1]]
        header_columns = [cell.column_letter for cell in sheet[1]]
        headers = dict(zip(header_columns, header_values))
        assert headers["E"].startswith("Buyer")


def test_pretty_headers_es():
    runner = CliRunner()
    with runner.isolated_filesystem():
        shutil.copyfile(FILENAME, "data.json")
        result = runner.invoke(cli, ["--human", "--language", "es", "data.json"])
        assert result.exit_code == 0
        path = pathlib.Path("result.xlsx")
        assert path.resolve().exists()
        xlsx_reader = openpyxl.load_workbook(path)
        sheet = xlsx_reader["parties"]
        header_values = [cell.value for cell in sheet[1]]
        header_columns = [cell.column_letter for cell in sheet[1]]
        headers = dict(zip(header_columns, header_values))
        assert headers["E"].startswith("Comprador")



from setuptools import find_packages, setup

with open("README.rst") as f:
    long_description = f.read()

setup(
    name="spoonbill",
    version="1.0.10",
    author="Open Contracting Partnership",
    author_email="data@open-contracting.org",
    url="https://github.com/open-contracting/spoonbill",
    description="Converts data that conforms to the Open Contracting Data Standard from JSON to Excel / CSV",
    license="BSD",
    packages=find_packages(exclude=["tests", "tests.*"]),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    install_requires=[
        "click",
        "click_logging",
        "flatten-dict",
        "ijson>=2.5",
        "jsonref",
        "ocdsextensionregistry",
        "ocdskit>=1.0.1",
        "requests",
        "scalpl",
        "setuptools",
        "xlsxwriter",
        'dataclasses;python_version<"3.7"',
    ],
    extras_require={
        "test": [
            "coveralls",
            "jsonpointer",
            "jmespath",
            "openpyxl",
            "pytest",
            "pytest-cov",
        ],
        "docs": [
            "furo",
            "sphinx",
            "sphinx-autobuild",
            "sphinx_click",
        ],
    },
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    entry_points={
        "console_scripts": [
            "spoonbill = spoonbill.cli:cli",
        ],
    },
)



import codecs
import copy
import functools
import gzip
import json
import logging
import re
from collections import OrderedDict
from itertools import chain
from numbers import Number
from pathlib import Path

import ijson
import jsonref
import requests
from scalpl import Cut

from spoonbill.common import COMBINED_TABLES, SEPARATOR

PYTHON_TO_JSON_TYPE = {
    "list": "array",
    "dict": "object",
    "string": "string",
    "int": "integer",
    "float": "number",
}
LOGGER = logging.getLogger("spoonbill")

ABBREVIATION_KEY = {
    "additionalIdentifiers": "ids",
    "additionalClassifications": "class",
    "documents": "docs",
}

ABBREVIATION_TABLE_NAME = {
    "contracts_implementation": "implementation",
    "contracts_implementation_transactions": "transactions",
}

GZIP_MAGIC_NUMBER = (b"\x1f", b"\x8b")


@functools.lru_cache(maxsize=None)
def common_prefix(path, subpath, separator="/"):
    """Given two paths, returns the longest common sub-path.

    >>> common_prefix('/contracts', '/contracts/items')
    '/contracts'
    >>> common_prefix('/tender/submissionMethod', '/tender/submissionMethodDetails')
    '/tender'
    >>> common_prefix('/tender/items/id', '/tender/items/description')
    '/tender/items'
    >>> common_prefix('/tender/items/0/additionalClassifications/0/id', '/tender/items/0')
    '/tender/items/0'
    """
    paths = [path.split(separator), subpath.split(separator)]
    if len(paths[0]) <= len(paths[1]):
        s1, s2 = paths
    else:
        s2, s1 = paths
    for i, path in enumerate(s1):
        if path != s2[i]:
            common = s1[:i]
            break
    else:
        common = s1
    return separator.join(common)


def iter_file(fd, root, multiple_values=False):
    """Iterate over `root` array in file provided by `filename` using ijson

    :param fd: File descriptor
    :param str root: Array field name inside file
    :param bool multiple_values: Determine line-delimited JSON
    :return: Iterator of bytes read and item as a tuple

    >>> [r for r in iter_file(open('tests/data/ocds-sample-data.json', 'rb'), 'records')]
    []
    >>> len([r for r in iter_file(open('tests/data/ocds-sample-data.json', 'rb'), 'releases')])
    6
    """
    reader = ijson.items(fd, prefix=("" if multiple_values else f"{root}.item"), multiple_values=multiple_values)
    for item in reader:
        yield item


def extract_type(item):
    """Extract item possible types from jsonschema definition.
    >>> extract_type({'type': 'string'})
    ['string']
    >>> extract_type(None)
    []
    >>> extract_type({})
    []
    >>> extract_type({'type': ['string', 'null']})
    ['string', 'null']
    """
    if not item or "type" not in item:
        return []
    type_ = item["type"]
    if not isinstance(type_, list):
        type_ = [type_]
    return type_


def validate_type(type_, item):
    """Validate if python object corresponds to provided type
    >>> validate_type(['string'], 'test_string')
    True
    >>> validate_type(['number'], 11.1)
    True
    >>> validate_type(['number'], 11)
    True
    >>> validate_type(['array'], [])
    True
    >>> validate_type(['array'], {})
    False
    >>> validate_type(['object'], [])
    False
    >>> validate_type(['object'], {})
    True
    """
    if isinstance(item, Number):
        name = "number"
    else:
        name = type(item).__name__
    expected = PYTHON_TO_JSON_TYPE.get(name)
    if expected:
        return expected in type_
    return True


def get_root(table):
    """Extract top level toot table of `table`"""
    while table.parent:
        table = table.parent
    return table


def combine_path(root, path, index="0", separator="/"):
    """Generates index based header for combined column"""
    combined_path = path
    for array in sorted(root.arrays, reverse=True):
        if common_prefix(path, array) == array:
            chunk = separator.join((array, index))
            combined_path = combined_path.replace(array, chunk)
    return combined_path


def get_matching_tables(tables, path):
    """Get list of matching tables for provided path

    Return list is sorted by longest matching path part

    :param tables: List of `Table' objects
    :param path: Path like string
    :return: List of matched by path tables
    """
    candidates = []
    for table in tables.values():
        for candidate in table.path:
            if common_prefix(candidate, path) == candidate:
                candidates.append(table)

    return sorted(candidates, key=lambda c: max((len(p) for p in c.path)), reverse=True)


def generate_table_name(parent_table, parent_key, key):
    """Generates name for non root table, to be used as sheet name

    :param str parent_table: Parent table name
    :param str parent_key: Parent object field name
    :param str key: Current object field name
    :return: Generated table name
    :rtype: str

    >>> generate_table_name('tenders', 'tender', 'items')
    'tenders_items'
    >>> generate_table_name('tenders', 'items', 'additionalClassifications')
    'tenders_items_class'
    >>> generate_table_name('parties', 'parties', 'roles')
    'parties_roles'
    """

    if key in ABBREVIATION_KEY:
        key = ABBREVIATION_KEY[key]

    if parent_key in parent_table:
        table_name = f"{parent_table}_{key}"
    else:
        table_name = f"{parent_table}_{parent_key}_{key}"

    if table_name in ABBREVIATION_TABLE_NAME:
        table_name = ABBREVIATION_TABLE_NAME[table_name]

    if len(table_name) >= 31:
        if parent_key in parent_table:
            table_name = f"{parent_table}_{key[:5]}"
        else:
            table_name = f"{parent_table}_{parent_key[:5]}_{key[:5]}"

    return table_name


def insert_after_key(columns, insert, last_key):
    data = {}
    for key, val in columns.items():
        data[key] = val
        if key == last_key:
            for k, v in insert.items():
                data[k] = v
    return data


def resolve_file_uri(file_path):
    """Read json file from provided uri
    :param file_path: URI to file, could be url or path
    :return: Read file as dictionary
    """
    if str(file_path).startswith("http://") or str(file_path).startswith("https://"):
        return requests.get(file_path).json()
    if isinstance(file_path, (str, Path)):
        with codecs.open(file_path, encoding="utf-8") as fd:
            return json.load(fd, object_pairs_hook=OrderedDict)


def read_lines(path):
    """Read file as lines"""
    with open(path, encoding="utf-8") as fd:
        return [line.strip() for line in fd.readlines()]


def get_pointer(table, abs_path, path, split, *, index=None):
    """Combine path and abs_path in order to fit table columns

    For example /tender/items/0/id should be /tender/items/0/id for tenders table
    but /tender/items/id for tenders_items table
    """
    array = table.is_array(path)
    if index and (array or table.is_combined):
        return SEPARATOR.join((abs_path, index))
    if table.is_root:
        return abs_path

    if array:
        paths = abs_path.split(SEPARATOR)
        prefix = ""

        for index, pth in enumerate(paths, 1):
            if pth.isdigit():
                continue
            if not pth:
                continue
            prefix = SEPARATOR.join((prefix, pth))
            if prefix == array:
                break
        pointer = SEPARATOR.join(paths[index:])
        if pointer:
            return SEPARATOR.join((prefix, pointer))
        return prefix
    return path


class RepeatFilter(logging.Filter):
    """
    Logger filter to avoid repeating of same messages during file processing
    """

    def filter(self, record):
        current_log = (record.module, record.levelno, record.msg)
        if current_log != getattr(self, "last_log", None):
            self.last_log = current_log
            return True
        return False


def make_count_column(array):
    """Make column name for arrays elements count

    >>> make_count_column('/tender/items')
    '/tender/itemsCount'
    >>> make_count_column('/tender/items/additionalClassifications')
    '/tender/items/additionalClassificationsCount'
    >>> make_count_column('/tender/items/')
    '/tender/itemsCount'
    """

    return array.rstrip("/") + "Count"


def get_reader(path):
    """
    Get reader function for a respective file format
    :param path: path to a file
    :return: reader function
    """
    with open(path, "rb") as f:
        first_bytes = f.read(2)
    if (first_bytes[0:1], first_bytes[1:2]) == GZIP_MAGIC_NUMBER:
        return gzip.open
    else:
        return open


def get_order(properties):
    order = [name for name in chain(properties, COMBINED_TABLES)]
    if "tender" in order:
        order[order.index("tender")] = "tenders"
    return order


def nonschema_title_formatter(title):
    """
    Formatting a path, that is absent in schema, to human-readable form
    :param title: str
    :return: formatted title

    >>> nonschema_title_formatter('legalEntityTypeDetail')
    'Legal Entity Type Detail'
    >>> nonschema_title_formatter('fuenteFinanciamiento')
    'Fuente Financiamiento'
    >>> nonschema_title_formatter('Óóó-Ñññ_Úúú')
    'Óóó Ñññ Úúú'
    """
    title = title.replace("_", " ").replace("-", " ")
    title = re.sub(r"(?<![A-Z])(?<!^)([A-Z])", r" \1", title)
    title = title.replace("  ", " ").replace("/", ": ")
    if title.startswith(": "):
        title = title[2:]
    title = title.title()
    return title


class SchemaHeaderExtractor:
    """
    Human-readable headers extracted from schema

    :param schema: The dataset's schema
    """

    def __init__(self, schema):
        self.schema = schema
        if not isinstance(self.schema, jsonref.JsonRef) and not isinstance(self.schema, OrderedDict):
            self.schema = jsonref.JsonRef.replace_refs(self.schema)

    def _get_header(self, id, paths):
        final_title = []
        for path in paths:
            _object = Cut(self.schema)["properties." + ".".join(path[:-1])]
            if hasattr(_object, "__reference__") and "title" in _object.__reference__:
                title = _object.__reference__["title"]
            else:
                title = Cut(self.schema)["properties." + ".".join(path)]
            if isinstance(title, dict):
                continue
            final_title.append(title)
        if id.startswith("/documents"):
            final_title = final_title[3:]
        if "Organization reference" in final_title:
            final_title.remove("Organization reference")
        return ": ".join(final_title)

    def get_header(self, id, paths):
        if paths and isinstance(paths, list):
            return self._get_header(id, paths)
        elif paths == []:
            return nonschema_title_formatter(id)
        else:
            return nonschema_title_formatter(paths)


def generate_paths(source):
    """
    Generate full paths for nested dictionary or alike object
    :param source: Input object
    :return: List of paths
    """
    paths = []
    if isinstance(source, dict):
        for k, v in source.items():
            paths.append([k])
            paths += [[k] + x for x in generate_paths(v)]
    return paths


def add_paths_to_schema(schema):
    """
    Extracts full path for each title in schema; creates paths list of full title for each
    :param unres_schema: Unresolved schema
    :param schema: Schema object that will be updated
    :return: Schema with full title paths
    """
    proxy = Cut(copy.deepcopy(schema))
    updated_items = {}
    for table in proxy["properties"].keys():
        path_item = Cut({table: copy.deepcopy(proxy["properties"][table])})
        # Generating path to each title in schema
        for path in generate_paths({table: proxy["properties"][table]}):
            if path[-1] == "title":
                object_path = ".".join(path[:-1])
                path_item[object_path]["$path"] = path
        updated_items[table] = path_item[table]
    proxy["properties"] = updated_items
    # Generating array with title paths for each title in schema
    title_paths = [path for path in title_path(proxy["properties"])]

    for path_list in title_paths:
        location = "properties." + ".".join(path_list[-1][:-1])
        proxy[location]["$title"] = path_list

    return proxy


def title_path(schema, path=[]):
    """
    Get path for each previous title and form full list of titles for each object
    :param schema: Object
    :param path: List of paths for full title
    :return:
    """
    for k, v in schema.items():
        newpath = path
        if "$path" in schema:
            newpath = path + [schema.get("$path", [])]
        if isinstance(v, dict):
            for result in title_path(v, newpath):
                if result:
                    yield result
        else:
            if newpath:
                yield newpath


def get_nestiness(abs_path):
    arrays = [c for c in abs_path if c.isdigit()]
    return len(arrays) - 1


def get_path_for_array_col(abs_path, array):
    nestiness = get_nestiness(abs_path)
    chunks = abs_path.split("/")[len(array.split("/")) + nestiness :]
    return "/".join(chain([array], chunks))



from spoonbill.cli import ANALYZED_LABEL
from spoonbill.i18n import translate


def test_translate_absent_locale():
    translation = translate(ANALYZED_LABEL, "uk_UA")
    assert translation == ANALYZED_LABEL



import csv
from pathlib import Path
from unittest.mock import call, patch

import openpyxl
from scalpl import Cut

from spoonbill import FileAnalyzer, FileFlattener
from spoonbill.flatten import Flattener, FlattenOptions
from spoonbill.utils import SchemaHeaderExtractor, add_paths_to_schema, generate_paths
from spoonbill.writers.csv import CSVWriter
from spoonbill.writers.xlsx import XlsxWriter

from .conftest import releases_extension_path, releases_path
from .utils import get_writers, prepare_tables, read_csv_headers, read_xlsx_headers

ID_FIELDS = {"tenders": "/tender/id", "parties": "/parties/id"}


def test_writer_init(spec, tmpdir, flatten_options):
    tables = prepare_tables(spec, flatten_options, ID_FIELDS)
    workdir = Path(tmpdir)
    get_writers(workdir, tables, flatten_options)
    path = workdir / "result.xlsx"

    assert path.is_file()
    for name in flatten_options.selection:
        path = workdir / f"{name}.csv"
        assert path.is_file()


def test_headers_filtering(spec, tmpdir, flatten_options):
    tables = prepare_tables(spec, flatten_options, ID_FIELDS)
    workdir = Path(tmpdir)
    get_writers(workdir, tables, flatten_options)
    for name in flatten_options.selection:
        path = workdir / f"{name}.csv"
        xlsx_headers = read_xlsx_headers(workdir / "result.xlsx", name)
        csv_headers = read_csv_headers(path)
        assert len(xlsx_headers) == 1
        assert len(csv_headers) == 1
        assert xlsx_headers[0] == ID_FIELDS[name]
        assert csv_headers[0] == ID_FIELDS[name]


def test_writers_pretty_headers(spec, tmpdir, releases, schema):
    # increase items count for force split
    releases[0]["tender"]["items"] = releases[0]["tender"]["items"] * 6
    for _ in spec.process_items(releases):
        pass
    options = FlattenOptions(
        **{
            "selection": {
                "tenders": {"split": True, "pretty_headers": True},
                "parties": {"split": False, "pretty_headers": True},
                "tenders_items": {
                    "split": True,
                    "headers": {"/tender/items/id": "item id"},
                    "pretty_headers": True,
                },
            }
        }
    )
    tables = {
        "tenders": spec.tables["tenders"],
        "parties": spec.tables["parties"],
        "tenders_items": spec.tables["tenders_items"],
    }
    tables_headers = {
        "tenders": [
            "Ocid",
            "Id",
            "Row Id",
            "Parent Id",
            "Tender: Award Criteria",
            "Tender: Award Criteria Details",
            "Tender: Tender Description",
            "Tender: Has Enquiries?",
            "Tender: Tender Id",
            "Tender: Main Procurement Category",
            "Tender: Number Of Tenderers",
            "Tender: Procurement Method",
            "Tender: Procurement Method Details",
            "Tender: Procurement Method Rationale",
            "Tender: Tender Status",
            "Tender: Submission Method",
            "Tender: Submission Method Details",
            "Tender: Tender Title",
            "Tender: Value: Amount",
            "Tender: Value: Currency",
            "Tender: Tender Period: Duration ( Days)",
            "Tender: Tender Period: End Date",
            "Tender: Tender Period: Start Date",
            "Tender: Tenderers: Organization Id",
            "Tender: Tenderers: Organization Name",
            "Tender: Value: Amount",
            "Tender: Value: Currency",
            "Tender: Tender Period: Duration ( Days)",
            "Tender: Tender Period: End Date",
            "Tender: Tender Period: Start Date",
            "Tender: Tender Period: Duration ( Days)",
            "Tender: Tender Period: End Date",
            "Tender: Tender Period: Start Date",
            "Tender: Tender Period: End Date",
            "Tender: Tender period: Start date",
        ],
        "parties": [
            "Ocid",
            "Id",
            "Row Id",
            "Parent Id",
            "Parties: Organization: Entity Id",
            "Parties: Organization: Common Name",
            "Parties: Organization: Party Roles",
            "Parties: Organization: Primary Identifier: Id",
            "Parties: Organization: Primary Identifier: Legal Name",
            "Parties: Organization: Primary Identifier: Scheme",
            "Parties: Organization: Contact Point: Email",
            "Parties: Organization: Contact Point: Fax Number",
            "Parties: Organization: Contact Point: Name",
            "Parties: Organization: Contact Point: Telephone",
            "Parties: Organization: Contact Point: Url",
            "Parties: Organization: Address: Country Name",
            "Parties: Organization: Address: Locality",
            "Parties: Organization: Address: Postal Code",
            "Parties: Organization: Address: Region",
            "Parties: Organization: Address: Street Address",
            "Buyer: Organization Id",
            "Buyer: Organization Name",
            "Parties: Test",
        ],
        "tenders_items": [
            "Ocid",
            "Id",
            "Row Id",
            "Parent Id",
            "Parent Table",
            "Tender: Items To Be Procured: Item: Description",
            "item id",
            "Tender: Items To Be Procured: Item: Quantity",
            "Tender: Items To Be Procured: Item: Unit: Id",
            "Tender: Items To Be Procured: Item: Unit: Name",
            "Tender: Items To Be Procured: Item: Unit: Scheme",
            "Tender: Value: Amount",
            "Tender: Value: Currency",
        ],
    }

    workdir = Path(tmpdir)
    get_writers(workdir, tables, options, schema)
    xlsx = workdir / "result.xlsx"

    for name, opts in options.selection.items():
        path = workdir / f"{name}.csv"
        xlsx_headers = read_xlsx_headers(xlsx, name)
        csv_headers = read_csv_headers(path)
        headers = tables_headers[name]
        assert not set(headers).difference(set(xlsx_headers))
        assert not set(headers).difference(set(csv_headers))

    options = FlattenOptions(
        **{
            "selection": {
                "tenders": {
                    "split": True,
                    "headers": {"/tender/id": "TENDER"},
                    "pretty_headers": True,
                },
                "tenders_items": {
                    "split": True,
                    "headers": {"/tender/items/id": "item id"},
                    "pretty_headers": True,
                },
                "parties": {
                    "split": False,
                    "headers": {"/parties/id": "PARTY"},
                    "pretty_headers": True,
                },
            }
        }
    )

    workdir = Path(tmpdir)
    get_writers(workdir, tables, options, schema)
    xlsx = workdir / "result.xlsx"

    sheet = "tenders"
    path = workdir / f"{sheet}.csv"
    xlsx_headers = read_xlsx_headers(xlsx, sheet)
    csv_headers = read_csv_headers(path)
    assert "TENDER" in xlsx_headers
    assert "TENDER" in csv_headers

    sheet = "tenders_items"
    path = workdir / f"{sheet}.csv"
    xlsx_headers = read_xlsx_headers(xlsx, sheet)
    csv_headers = read_csv_headers(path)
    assert "item id" in xlsx_headers
    assert "item id" in csv_headers

    xlsx = workdir / "result.xlsx"
    sheet = "parties"
    path = workdir / f"{sheet}.csv"

    xlsx_headers = read_xlsx_headers(xlsx, sheet)
    csv_headers = read_csv_headers(path)

    assert "PARTY" in xlsx_headers
    assert "PARTY" in csv_headers


def test_writers_flatten_count(spec, tmpdir, releases, schema):
    releases[0]["tender"]["items"] = releases[0]["tender"]["items"] * 6
    for _ in spec.process_items(releases):
        pass
    options = FlattenOptions(
        **{
            "selection": {
                "tenders": {"split": True, "pretty_headers": True},
                "parties": {"split": True, "pretty_headers": True},
                "tenders_items": {
                    "split": False,
                    "pretty_headers": True,
                },
                "parties_ids": {
                    "split": False,
                    "pretty_headers": True,
                },
            },
            "count": True,
        }
    )

    workdir = Path(tmpdir)
    analyzer = FileAnalyzer(workdir)
    flattener = FileFlattener(
        workdir=workdir, options=options, tables=spec.tables, csv=True, analyzer=analyzer, schema=schema
    )
    xlsx = workdir / "result.xlsx"
    for _ in flattener.flatten_file(releases_path):
        pass
    sheet = "tenders"
    path = workdir / f"{sheet}.csv"
    # TODO: xlsx headers
    headers = read_csv_headers(path)
    assert "Tender: Items Count" in headers
    assert "Tender: Tenderers Count" in headers

    sheet = "parties"
    path = workdir / f"{sheet}.csv"
    for headers in read_xlsx_headers(xlsx, sheet), read_csv_headers(path):
        assert "Parties: Additional Identifiers Count" in headers


def test_writers_table_name_override(spec, tmpdir, schema):
    options = FlattenOptions(
        **{
            "selection": {
                "parties": {"split": False, "pretty_headers": True, "name": "testname"},
                "tenders": {"split": True, "pretty_headers": True, "name": "my_tenders"},
                "tenders_items": {"split": False, "pretty_headers": True, "name": "pretty_items"},
            }
        }
    )
    tables = prepare_tables(spec, options)
    for name, table in tables.items():
        for col in table:
            table.inc_column(col, col)

    workdir = Path(tmpdir)
    get_writers(workdir, tables, options, schema)
    xlsx = workdir / "result.xlsx"
    for name in ("testname", "my_tenders", "pretty_items"):
        path = workdir / f"{name}.csv"
        assert path.is_file()
        assert read_xlsx_headers(xlsx, name)
        assert read_csv_headers(path)


def test_abbreviations(spec, tmpdir):
    options = FlattenOptions(
        **{
            "selection": {
                "tenders_items_class": {"split": False},
                "parties_ids": {"split": False},
                "transactions": {"split": False},
            }
        }
    )
    new_names = ["tenders_items_class", "parties_ids", "transactions"]
    tables = prepare_tables(spec, options)
    for name, table in tables.items():
        for col in table:
            table.inc_column(col, col)

    workdir = Path(tmpdir)
    get_writers(workdir, tables, options)
    xlsx = workdir / "result.xlsx"
    for name in new_names:
        path = workdir / f"{name}.csv"
        assert path.is_file()
        assert read_xlsx_headers(xlsx, name)
        assert read_csv_headers(path)


def test_name_duplicate(spec, tmpdir, schema):
    duplicate_name = "test"
    options = FlattenOptions(
        **{
            "selection": {
                "parties": {"split": False, "pretty_headers": True, "name": duplicate_name},
                "tenders": {"split": True, "pretty_headers": True, "name": duplicate_name},
                "tenders_items": {"split": False, "pretty_headers": True, "name": duplicate_name},
            }
        }
    )
    tables = prepare_tables(spec, options)
    for name, table in tables.items():
        for col in table:
            table.inc_column(col, col)
    workdir = Path(tmpdir)
    get_writers(workdir, tables, options, schema)
    xlsx = workdir / "result.xlsx"
    for name in ("test", "test1", "test2"):
        path = workdir / f"{name}.csv"
        assert path.is_file()
        assert read_xlsx_headers(xlsx, name)
        assert read_csv_headers(path)


@patch("spoonbill.LOGGER.error")
def test_writers_invalid_table(log, spec, tmpdir, schema):
    options = FlattenOptions(
        **{
            "selection": {
                "parties": {"split": False, "pretty_headers": True, "name": "testname"},
            }
        }
    )
    tables = prepare_tables(spec, options)
    for name, table in tables.items():
        for col in table:
            table.inc_column(col, col)

    workdir = Path(tmpdir)
    writers = get_writers(workdir, tables, options, schema)
    for writer in writers:
        writer.writerow("test", {})
    log.assert_has_calls([call("Invalid table test"), call("Invalid table test")])


@patch("spoonbill.LOGGER.error")
def test_writers_invalid_row(log, spec, tmpdir, schema):
    options = FlattenOptions(
        **{
            "selection": {
                "parties": {"split": False, "pretty_headers": True, "name": "testname"},
            }
        }
    )
    tables = prepare_tables(spec, options)
    for name, table in tables.items():
        for col in table:
            table.inc_column(col, col)

    workdir = Path(tmpdir)
    writers = get_writers(workdir, tables, options, schema)
    for writer in writers:
        writer.writerow("parties", {"/test/test": "test"})
    log.assert_has_calls(
        [
            call("Operation produced invalid path. This a software bug, please send issue to developers"),
            call("Failed to write row None with error dict contains fields not in fieldnames: '/test/test'"),
            call("Operation produced invalid path. This a software bug, please send issue to developers"),
            call("Failed to write column /test/test to xlsx sheet parties"),
        ]
    )


@patch("spoonbill.LOGGER.error")
@patch("spoonbill.writers.csv.open", **{"side_effect": OSError("test")})
def test_writers_open_fail(open_, log, spec, tmpdir):
    options = FlattenOptions(
        **{
            "selection": {
                "parties": {"split": False, "pretty_headers": True, "name": "testname"},
            }
        }
    )
    workdir = Path(tmpdir)
    tables = prepare_tables(spec, options)
    get_writers(workdir, tables, options)
    log.assert_has_calls([call(f"Failed to open file {tmpdir / 'testname.csv'} with error test")])


def test_csv_writer(spec_analyzed, releases, flatten_options, tmpdir, schema):
    flattener = Flattener(flatten_options, spec_analyzed.tables)
    flatten_options.selection["parties"].split = True
    tables = prepare_tables(spec_analyzed, flatten_options)
    workdir = Path(tmpdir)
    with CSVWriter(workdir, tables, flatten_options, schema) as writer:
        # Writing CSV files
        for _count, flat in flattener.flatten(releases):
            for name, rows in flat.items():
                for row in rows:
                    writer.writerow(name, row)

    # Reading CSV files
    counter = {}
    for _count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            if name not in counter:
                counter[name] = 0
            for row in rows:
                str_row = {k: str(v) for (k, v) in row.items()}
                file = name + ".csv"
                path = workdir / file
                with open(path, newline="", encoding="utf-8") as csv_file:
                    csv_reader = csv.DictReader(csv_file)
                    for num, line in enumerate(csv_reader):
                        if num == counter[name]:
                            clean_line = {k: v for (k, v) in line.items() if v != ""}
                            assert dict(clean_line) == str_row
                counter[name] += 1


def test_xlsx_writer(spec_analyzed, releases, flatten_options, tmpdir, schema):
    flattener = Flattener(flatten_options, spec_analyzed.tables)
    tables = prepare_tables(spec_analyzed, flatten_options)
    workdir = Path(tmpdir)
    with XlsxWriter(workdir, tables, flatten_options, schema) as writer:
        # Writing XLSX file
        for _count, flat in flattener.flatten(releases):
            for name, rows in flat.items():
                for row in rows:
                    writer.writerow(name, row)

    # Reading XLSX files
    counter = {}
    path = workdir / "result.xlsx"
    xlsx_reader = openpyxl.load_workbook(path)

    for _count, flat in flattener.flatten(releases):
        for name, rows in flat.items():
            if name not in counter:
                counter[name] = 2
            sheet = xlsx_reader[name]
            headers = {cell.column_letter: cell.value for cell in sheet[1]}
            for row in rows:
                line = {headers[cell.column_letter]: cell.value for cell in sheet[counter[name]]}
                row = row.as_dict()
                assert not set(row.keys()).difference(set(line.keys()))
                for k, v in row.items():
                    assert str(v) == str(line[k])
                counter[name] += 1


def test_xlsx_only_no_default_columns(spec_analyzed, releases, tmpdir, schema):
    flatten_options = FlattenOptions(**{"selection": {"tenders": {"split": True, "only": ["/tender/id"]}}})
    flattener = Flattener(flatten_options, spec_analyzed.tables)
    tables = prepare_tables(spec_analyzed, flatten_options)
    workdir = Path(tmpdir)
    with XlsxWriter(workdir, tables, flatten_options, schema) as writer:
        for _count, flat in flattener.flatten(releases):
            for name, rows in flat.items():
                for row in rows:
                    writer.writerow(name, row)

    path = workdir / "result.xlsx"
    xlsx_reader = openpyxl.load_workbook(path)
    column = []
    for row in xlsx_reader["tenders"].rows:
        column.append(row[0].value)
    assert column[0] == "/tender/id"
    assert xlsx_reader["tenders"].max_column == 1


def test_flatten_multiple_files(spec, tmpdir, releases, schema):
    for _ in spec.process_items(releases):
        pass
    options = FlattenOptions(**{"selection": {"tenders": {"split": True}}})

    workdir = Path(tmpdir)
    analyzer = FileAnalyzer(workdir)
    flattener = FileFlattener(
        workdir=workdir, options=options, tables=spec.tables, csv=True, analyzer=analyzer, schema=schema
    )
    xlsx = workdir / "result.xlsx"
    sheet = "tenders"
    for _ in flattener.flatten_file(releases_path):
        pass
    wb = openpyxl.load_workbook(xlsx)
    ws = wb[sheet]
    line_number = ws.max_row - 1
    assert ws.max_row - 1 == 4

    flattener = FileFlattener(
        workdir=workdir, options=options, tables=spec.tables, csv=True, analyzer=analyzer, schema=schema
    )
    for _ in flattener.flatten_file([releases_path, releases_path]):
        pass
    wb = openpyxl.load_workbook(xlsx)
    ws = wb[sheet]
    assert ws.max_row - 1 == line_number * 2


def test_extension_export(spec, tmpdir, releases_extension, schema):
    for _ in spec.process_items(releases_extension):
        pass
    options = FlattenOptions(**{"selection": {"tenders": {"split": False}, "documents": {"split": False}}})

    workdir = Path(tmpdir)
    analyzer = FileAnalyzer(workdir)
    flattener = FileFlattener(workdir=workdir, options=options, tables=spec.tables, analyzer=analyzer, schema=schema)
    xlsx = workdir / "result.xlsx"
    sheet = "documents"
    extension_header = "/documents/test_extension"
    for _ in flattener.flatten_file(releases_extension_path):
        pass
    wb = openpyxl.load_workbook(xlsx)
    ws = wb[sheet]
    for column_cell in ws.iter_cols(1, ws.max_column):
        if column_cell[0].value == extension_header:
            extension_column = ws[column_cell[0].coordinate[:-1]]
    for cell in extension_column:
        if cell.value == extension_header:
            continue
        assert cell.value == "test"
    for column_cell in wb["tenders"].iter_cols(1, ws.max_column):
        assert column_cell[0].value != extension_header


def test_schema_header_paths(schema):
    paths = generate_paths(schema["properties"])
    schema = add_paths_to_schema(schema)
    proxy = Cut(schema["properties"])
    for path in paths:
        if path[-1] == "title":
            _path = ".".join(path[:-1])
            assert "$title" in proxy[_path]
            assert proxy[_path]["$title"][-1] == path


def test_schema_header_generation(schema):
    headers = SchemaHeaderExtractor(schema)
    paths = [
        ["parties", "title"],
        ["parties", "items", "title"],
        ["parties", "items", "properties", "contactPoint", "title"],
        ["parties", "items", "properties", "contactPoint", "properties", "name", "title"],
    ]
    assert headers.get_header("", paths) == "Parties: Organization: Contact point: Name"
